using CHNw.CommonLib.Core;                        // for RandomSource, _
using CHNw.CommonLib.Data;                        // for RichSqlCommand(4), RichDataTable(3)
using CHNw.CommonLib.Trace;                       // for ExceptionAction(2)
using NUnit.Framework;
using NUnit.Framework.SyntaxHelpers;
using System;                                     // for Guid(3)
using System.Collections.Generic;                 // for List<>
using System.Data;                                // for CommandType(2)
using System.Data.SqlClient;                      // for SqlDataReader(1)
using System.Data.SqlTypes;                       // for SqlBoolean(3)

namespace DataWarehouse.LibTests
{
    // ------------------------------------------------------------------------------------------100
    /// <!-- DataTests -->
    /// <summary>
    ///      The DataTests class tests the classes in the Data Library
    /// </summary>
    [TestFixture]
    public class DataTests
    {
        private string _result;
        private static string test1 = "\r\n((((((((((   ";
        private static string test2 = " test   ))))))))))\r\n";


        // --------------------------------------------------------------------------------------100
        /// <!-- AllTests -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string AllTests()
        {
            _result = "";
            string result = "";


            string b = SqlBoolean.True.ToString();
            b = SqlBoolean.False.ToString();
            b = SqlBoolean.Null.ToString();
            _result += b;


            RichDataTable_RichSqlCommand_test();   result += _.Affix(test1 + "RichDataTable"     + test2, _result);
            RichDataTable_Add_test();              result += _.Affix(test1 + "RDT Add(copy) row" + test2, _result);
            RichDataTable_Copy_test();             result += _.Affix(test1 + "RDT Copy table"    + test2, _result);
            RichDataTable_Distinct_test();         result += _.Affix(test1 + "RDT Distinct"      + test2, _result);
            RichDataTable_Split_test();            result += _.Affix(test1 + "RDT Split"         + test2, _result);
            //RichDataTable_XmlField_test();         result += _.Affix(test1 + "RDT XML field"     + test2, _result);
            RichSqlCommand_ExecuteReader_test();   result += _.Affix(test1 + "RSC ExecuteReader" + test2, _result);


            result += "\r\n" + "test done";
            return result;
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_Split_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichDataTable_Split_test()
        {
            RichDataTable rt = SimpleTestTable();
            List<RichDataTable> table = rt._Split("Age");
            Assert.That(table.Count == 2);
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_Distinct_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichDataTable_Distinct_test()
        {
            _result = "";
            RichDataTable rt = SimpleTestTable();
            List<object> list = rt._Distinct("Age");
            Assert.That(list.Count == 2);
            _result = "OK";
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_Add_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichDataTable_Add_test()
        {
            RichDataTable rt = SimpleTestTable();
            RichDataTable rt2 = rt.Clone();
            Assert.That(rt2.Columns[1].ColumnName == rt.Columns[1].ColumnName);
            rt2.Add(rt, 1); // _CopyRowToTable(1, rt2);
            rt2.Add(rt, 0); // _CopyRowToTable(0, rt2);
            int age1 = rt._Int(1, 1, 1);
            int age2 = rt2._Int(0, 1, 2);
            Assert.That(age2 == age1);
            string name1 = rt._Str(0, 0, "a");
            string name2 = rt2._Str(1, 0, "b");
            Assert.That(name2 == name1);
            _result = "OK";
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_Copy_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichDataTable_Copy_test()
        {
            RichDataTable rt = SimpleTestTable();
            RichDataTable rt2 = rt.Copy();
            int row = rt.Count - 1;
            Assert.That(rt2._Str(row, 0, "b") == rt2._Str(row, 0, "a"));
            Assert.That(rt2._Int(row, 1, 1)   == rt2._Int(row, 1, 0));
            _result = "OK";
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_RichSqlCommand_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichDataTable_RichSqlCommand_test()
        {
            _result = "";
            RichSqlCommand cmd = new RichSqlCommand(CommandType.Text, DBrules.Conn
                , "SELECT * FROM Star.Facility"
                , ExceptionAction.Squawk);
            RichDataTable facility = new RichDataTable(cmd, "Facility");
            Assert.That(facility.Columns.Count > 0);
            _result = "OK";
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- RichDataTable_XmlField_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        //public void RichDataTable_XmlField_test()
        //{
        //    _result = "";
        //    // -------------------------------------------------------------------------90
        //    //  Test the _XmlField method
        //    // -------------------------------------------------------------------------90
        //    Guid patientGuid = new Guid("e5dd0aff-e5b5-4d2e-b476-00fb50f59f04"); // An A08
        //    RichDataTable patient = (new HL7Procedure()).Test_ADTrecord_Read(patientGuid);
        //    Assert.That(string.IsNullOrEmpty(patient.Errors));
        //    string xml = patient._Str(0, "ADTrecord", "");
        //    Assert.That(!string.IsNullOrEmpty(xml));
        //    string patientVisitNumber = patient._XmlStr(0, "ADTrecord", "", CommonXpath.VisitNumber);
        //    Assert.That(patientVisitNumber == "0830901493");
        //    _result = "OK";
        //}


        // --------------------------------------------------------------------------------------100
        /// <!-- RichSqlCommand_ExecuteReader_test -->
        /// <summary>
        /// 
        /// </summary>
        [Test]
        public void RichSqlCommand_ExecuteReader_test()
        {
            _result = "";
            RichSqlCommand cmd = new RichSqlCommand(CommandType.Text, DBrules.Conn
                , "SELECT * FROM Star.Facility"
                , ExceptionAction.Squawk);
            cmd._Open();
            cmd._StatementCompleted = false;
            SqlDataReader dr = cmd.ExecuteReader();
            Assert.That(dr.FieldCount > 0);
            cmd.Connection.Close();
            _result = "OK";
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- SimpleTestTable -->
        /// <summary>
        ///      Returns a simple two column thre row table for testing various Data methods
        /// </summary>
        /// <returns></returns>
        private static RichDataTable SimpleTestTable()
        {
            RichDataTable rt = new RichDataTable();
            rt._AddColumn("Name", typeof(string));
            rt._AddColumn("Age",  typeof(int));
            int row;
            row = rt._AddBlankRow(); rt.Rows[row]["Name"] = "Fred";    rt.Rows[row]["Age"] = 26;
            row = rt._AddBlankRow(); rt.Rows[row]["Name"] = "Jon";     rt.Rows[row]["Age"] = 48;
            row = rt._AddBlankRow(); rt.Rows[row]["Name"] = "Melinda"; rt.Rows[row]["Age"] = 26;
            return rt;
        }


    }
}

