// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using InformationLib.DataAccess;
using InformationLib.Endemes;                     // for Endeme(1)
using InformationLib.SoftData;                    // for Tally(1)
using InformationLib.Strings;                     // for IResults(1), Date(3)
using System;                                     // for many(8)
using System.Collections;                         // for ArrayList
using System.Collections.Generic;                 // for List(1)
using System.Data;                                // for many(10)
using System.Data.SqlClient;                      // for SqlDataReader(3)
using System.Data.SqlTypes;                       // for SqlDateTime(16), SqlXml(1)
using System.Reflection;                          // for GetProperties
using System.Text;                                // for StringBuilder(2)
using System.Text.RegularExpressions;             // for Regex(3), MatchCollection(1)
using System.Xml;                                 // for XmlNode(2), XmlNodeList(1)
using System.Data.OleDb;

namespace InformationLib.DataAccess
{
    // --------------------------------------------------------------------------------------------
    /// <!-- RichDataTable -->
    /// <summary>
    ///      The RichDataTable class is a DataTable class with a few added methods
    /// </summary>
    public class RichDataTable
    {

        #region members and constructors


        // ----------------------------------------------------------------------------------------
        /// <!-- Table -->
        private DataTable _table;
        /// <summary>
        /// 
        /// </summary>
        public DataTable Table
        {
            get { return _table; }
        }


        // ----------------------------------------------------------------------------------------
        //  Data members for handling errors
        // ----------------------------------------------------------------------------------------
        public string Errors  { get { return _errors;  } set { _errors = value; } }  private string _errors;  /// <summary>Records non-error results</summary>
        public string Results { get { return _results; } }                           private string _results;


        // ----------------------------------------------------------------------------------------
        /// <!-- LastRow, LastColumn, LastXdoc, LastXml -->
        private int             LastRow    { get { return _lastRow;    } set { _lastRow    = value; } } private int             _lastRow;
        private string          LastColumn { get { return _lastColumn; } set { _lastColumn = value; } } private string          _lastColumn;
        public  RichXmlDocument LastXdoc   { get { return _lastXdoc;   } set { _lastXdoc   = value; } } private RichXmlDocument _lastXdoc;
        private string          LastXml    { get { return _lastXml;    } set { _lastXml    = value; } } private string          _lastXml;
        private string p;
        private string p_2;


        // ----------------------------------------------------------------------------------------
        //  Simple Constructors
        // ----------------------------------------------------------------------------------------
        public RichDataTable()                      { _table = new DataTable(); _errors = ""; }
        public RichDataTable(string tableName)      { _table = new DataTable(tableName); _errors = ""; }
        public RichDataTable(DataTable dt)
        { if (dt == null)  _table = new DataTable(); else _table = dt; _errors = ""; }
        //public RichDataTable(DataTable dt, bool disposeSourceTable) // ignore boolean value
        //{ if (dt == null)  _table = new DataTable(); else _table = dt; _errors = ""; }


        // ----------------------------------------------------------------------------------------
        /// <!-- RichDataTable constructor -->
        /// <summary>
        ///      Constructor from a hash
        /// </summary>
        /// <param name="sid"></param>
        /// <param name="keyColumn"></param>
        /// <param name="valColumn"></param>
        public RichDataTable(Dictionary<string, int> sid, string keyColumn, string valColumn)
        {
            _table = new DataTable();
            _table.Columns.Add(keyColumn, typeof(string));
            _table.Columns.Add(valColumn, typeof(int));
            foreach (string key in sid.Keys)
            {
                int val = sid[key];
                int row = _table.Rows.Count;
                _table.Rows.Add(NewRow());
                _table.Rows[row][keyColumn] = key;
                _table.Rows[row][valColumn] = val;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RichDataTable constructor -->
        /// <summary>
        ///      Constructs a RichDataTable from a "SELECT" stored procedure or text command
        /// </summary>
        /// <param name="cmd">A command that returns a table as per a SELECT command</param>
        /// <param name="tableName">What you want to name the table that is returned</param>
        public RichDataTable(RichSqlCommand cmd, string tableName)
        {
            // -------------------------------------------------------------------------90
            //  Preprocessing
            // -------------------------------------------------------------------------90
            _table = new DataTable();
            _errors = "";
            cmd._Open();
            cmd._StatementCompleted = false;


            // -------------------------------------------------------------------------90
            //  Run the reader
            // -------------------------------------------------------------------------90
            //MessageBox.Show("Before execute reader");
            SqlDataReader dr = cmd.ExecuteReader();
            //MessageBox.Show("After execute reader");
            _errors  += _.Affix("\r\n", cmd.Errors);
            _results += _.Affix("\r\n", cmd.Results);


            // -------------------------------------------------------------------------90
            //  Decant the reader
            // -------------------------------------------------------------------------90
            if (dr != null)
                try
                {
                    dr = _FillFrom(dr, false);
                    _table.TableName = tableName;
                    cmd._StatementCompleted = true;
                    _results += _.Affix("\r\n", cmd.Results);
                }
                // ---------------------------------------------------------------------90
                //  Error handling
                // ---------------------------------------------------------------------90
                catch (Exception ex) {  _errors += "\r\n" + "RichDataTable constructor error - \r\n" + ex.Message; }
                finally
                {
                    _errors  += cmd.Errors;
                    _results += cmd.Results;
                    cmd.Connection.Close();
                    dr.Dispose();
                    dr = null;
                }
            else
            {
                cmd.Connection.Close();
                _errors += "\r\n" + "RichDataTable constructor error - \r\n" + cmd.Errors;
            }


            // -------------------------------------------------------------------------90
            //  Post error processing
            // -------------------------------------------------------------------------90
            if (!cmd._StatementCompleted)
                _errors += "\r\n" + "Error - statement did not complete:" + "\r\n" + cmd.CommandText;
            if (!string.IsNullOrEmpty(_errors))
                Throws.A(new DataException(_errors), Throws.Actions, "P");
        }

        
        // ----------------------------------------------------------------------------------------
        /// <!-- RichDataTable constructor -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="dr"></param>
        /// <param name="disposeSourceReader"></param>
        public RichDataTable(SqlDataReader dr, bool disposeSourceReader) : base()
        {
            dr = _FillFrom(dr, disposeSourceReader);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RichDataTable constructor -->
        /// <summary>
        ///      Reads from a file into a data table, creating all string columns
        /// </summary>
        /// <param name="path"></param>
        /// <param name="name"></param>
        public RichDataTable(string path, string name, List<string> columns)
        {
            Init(name, columns);
            List<string> str = _.GetListFromFile(path);
            for (int i = 0; i < str.Count; ++i)
                AddLineAsRow(str[i]);
        }


        // --------------------------------------------------------------------------------------
        /// <!-- AddLine -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="line"></param>
        private void AddLineAsRow(string line)
        {
            char[] sep = { '\t' };
            int row = _AddBlankRow();
            string[] cell = line.Split(sep);
            int count = Math.Min(cell.Length, _table.Columns.Count);
            for (int col = 0; col < count; ++col)
                _table.Rows[row][col] = cell[col];
        }


        // --------------------------------------------------------------------------------------
        /// <!-- Init -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="name"></param>
        /// <param name="columns"></param>
        private void Init(string name, List<string> columns)
        {
            _table = new DataTable(name);
            foreach (string column in columns)
                _AddColumn(column, typeof(string));
        }


        #endregion members and constructors

        #region methods and properties


        // ----------------------------------------------------------------------------------------
        //  Pass-through methods
        // ----------------------------------------------------------------------------------------
        /// <summary>Returns a new blank table with the same structure as this one</summary>
        public RichDataTable Clone   ()              { return (new RichDataTable(_table.Clone())); }
        public DataRow       NewRow  ()              { return _table.NewRow();                     }
        public void          RemoveAt(int row)       { _table.Rows.RemoveAt(row);                  }
        public DataRow[]     Select  (string filter) { return _table.Select(Regex.Replace(filter, "^ *where +", "", RegexOptions.IgnoreCase)); }


        // ----------------------------------------------------------------------------------------
        //  Pass-through properties
        // ----------------------------------------------------------------------------------------
        public DataColumnCollection Columns   { get { if (_table == null) return null;            else return _table.Columns;    } }  /// <summary>Counts the number of rows in the table</summary><returns>the number of rows in the table</returns>
        public int                  Count     { get { if (_table == null) return 0;               else return _table.Rows.Count; } }
        public DataRowCollection    Rows      { get { if (_table == null) return null;            else return _table.Rows;       } }
        public string               TableName { get { if (_table == null) return "table is null"; else return _table.TableName;  } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Add -->
        /// <summary>
        ///      Adds a row to the table (a real copy not a reference) from the passed table,
        ///      tables must have the same number of columns with the same types in the same order
        /// </summary>
        /// <param name="rt">table with the same type signature</param>
        /// <param name="row">row number to add</param>
        public void Add(RichDataTable rt, int row)
        {
            if (0 <= row && row < rt.Count)
            {
                DataRow copy = this.NewRow();
                DataRow from = rt.Rows[row];
                for (int i = 0; i < from.ItemArray.Length; ++i)
                    copy[i] = from[i];
                _table.Rows.Add(copy);
            }
        }
        public int Add(DataRow newRow)
        {
            int row = _table.Rows.Count;
            _table.Rows.Add(newRow);
            return row;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _AddColumn -->
        /// <summary>
        ///      Adds a column to the table
        /// </summary>
        /// <param name="column"></param>
        /// <param name="type"></param>
        /// <returns>true if the column was created, false if it was already there</returns>
        public bool _AddColumn(string column, Type type)
        {
            if (this.Table.Columns.Contains(column))
            {
                if (this.Table.Columns[column].GetType() != type)
                {
                    Throws.A(new NotSupportedException("Error -"
                        + " attempt to add a column that alread exists"
                        + " as a different type"
                        + " to the " + this.Table.TableName + " table"), Throws.Actions, "P");
                }
                return false;
            }
            else
            {
                this.Table.Columns.Add(column, type);
                return true;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _AddBlankRow -->
        /// <summary>
        ///      Adds a row to the table, returning the index of the row added
        /// </summary>
        /// <returns>the index of the added row</returns>
        public int _AddBlankRow()
        {
            int r = Count;
            DataRow row = NewRow();
            Rows.Add(row);
            return r;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- AllAvgs -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public RichDataTable AllAvgs()
        {
            RichDataTable avg = this.Clone();
            avg.Rows.Add(avg.NewRow());
            double count = this.Count;
            foreach (DataColumn column in avg.Columns)
            {
                Type type = column.DataType;
                switch (type.Name)
                {
                    case "Int16"   : avg.Rows[0][column] = (Int16)(this._Sum(column.ColumnName, 0) / count);               break;
                    case "String"  : avg.Rows[0][column] = "1";                                                            break;
                    case "DateTime": avg.Rows[0][column] = this._Average(column.ColumnName, Date.MinSqlValue.CLRFormat);   break;
                    case "Int32"   : avg.Rows[0][column] = (int)((0.5 + 1.0 * this._Sum(column.ColumnName, 0.0)) / count); break;
                    case "Double"  : avg.Rows[0][column] = this._Sum(column.ColumnName, 0.0) / count;                      break;
                    case "Single"  : avg.Rows[0][column] = this._Sum(column.ColumnName, 0.0F) / (float)count;              break;
                    default: break;
                }
            }

            return avg;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- AllSums -->
        /// <summary>
        ///      Returns the sums of all the columns
        /// </summary>
        /// <returns></returns>
        public RichDataTable AllSums()
        {
            RichDataTable sum = this.Clone();
            sum.Rows.Add(sum.NewRow());
            foreach (DataColumn column in sum.Columns)
            {
                Type type = column.DataType;
                switch (type.Name)
                {
                    case "Int16"   : sum.Rows[0][column] = (Int16)this._Sum(column.ColumnName, 0);   break;
                    case "String"  : sum.Rows[0][column] = "" + this.Count;                          break;
                    case "DateTime": sum.Rows[0][column] = this._Max(column.ColumnName, Date.MinSqlValue.CLRFormat); break;
                    case "Int32"   : sum.Rows[0][column] = this._Sum(column.ColumnName, 0);          break;
                    case "Double"  : sum.Rows[0][column] = this._Sum(column.ColumnName, 0.0);        break;
                    case "Single"  : sum.Rows[0][column] = this._Sum(column.ColumnName, 0.0F);       break;
                    default: break;
                }
            }

            return sum;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ArrayList -->
        /// <summary>
        ///      Copies a table column into an ArrayList of strings
        /// </summary>
        /// <param name="table"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        public ArrayList _ArrayList(string column)
        {
            ArrayList arr;
            if (_table != null)
            {
                int count = Count;
                arr = new ArrayList(count);


                // ---------------------------------------------------------------------90
                //  Add column values from DataTable to ArrayList
                // ---------------------------------------------------------------------90
                for (int row = 0; row < count; ++row)
                {
                    object obj = _table.Rows[row][column];
                    if (obj != null)
                        arr.Add(obj.ToString());
                }
            }
            else
                arr = new ArrayList();


            return arr;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ToList -->
        /// <summary>
        ///      Returns a table column as a List&lt;strings&gt;
        /// </summary>
        /// <param name="table"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        public List<string> _ToList(string column)
        {
            List<string> arr = new List<string>(Count);
            if (_table != null)
            {
                for (int row = 0; row < Count; ++row)
                {
                    object obj = _table.Rows[row][column];
                    if (obj != null)
                        arr.Add(obj.ToString());
                }
            }


            return arr;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Average -->
        /// <summary>
        ///      This is wrong but I'll use it for now as a stub
        /// </summary>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public DateTime _Average(string column, DateTime defaultValue)
        {
            //  This is wrong but I'll use it for now as a stub
            DateTime min = _Min(column, defaultValue);
            DateTime max = _Max(column, defaultValue);
            TimeSpan t = max - min;
            long half = t.Ticks / 2;
            DateTime mid = min.AddTicks(half);
            return mid;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Bool -->
        /// <summary>
        ///      Returns an exception-free boolean representation of the cell value
        /// </summary>
        /// <param name="row">cell row</param>
        /// <param name="column">cell column</param>
        /// <param name="defaultValue">value returned if the method can't figure it out</param>
        /// <returns>boolean representation of the cell value or default value</returns>
        public bool _Bool(int row, string column, bool defaultValue)
        {
            bool b;
            if (this._Exists(row, column))
                b = TreatAs.BoolValue(_table.Rows[row][column], defaultValue);
            else b = defaultValue;
            this._RecordLast(row, column);
            return b;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Char -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public char _Char(int row, string column, char defaultValue)
        {
            string str = (_Str(row, column, defaultValue.ToString()));
            return str.ToCharArray()[0];
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _CheckKeyUniqueness -->
        /// <summary>
        ///      Returns a message saying which single column keys are duplicated
        /// </summary>
        /// <param name="table"></param>
        /// <param name="keyColumn"></param>
        /// <param name="dt"></param>
        /// <returns></returns>
        public string _CheckKeyUniqueness(string keyColumn)
        {
            List<string> array = new List<string>();
            int count = Count;
            string table = _table.TableName;
            string delim = "";
            string str = "";


            for (int row = 0; row < count; ++row)
            {
                // ---------------------------------------------------------------------90
                //  Look to see if row key as already in array
                // ---------------------------------------------------------------------90
                string key = this._Str(row, keyColumn, "");
                if (array.Contains(key))
                {
                    str += delim + table + ": " + key + " not unique";
                    delim = "\r\n";
                }


                // ---------------------------------------------------------------------90
                //  Put row key in array
                // ---------------------------------------------------------------------90
                array.Add(key);
            }


            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Collate -->
        /// <summary>
        ///      Collates a column's values into a big long string
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public string _Collate(string column, string join)
        {
            StringBuilder str = new StringBuilder();
            string delim = "";
            int count = Rows.Count;
            for (int row = 0; row < count; ++row)
            {
                str.Append(delim + _Str(row, column, ""));
                delim = join;
            }
            return str.ToString();
        }
        public string _Collate(string column1, string column2, string colJoin, string rowJoin)
        {
            StringBuilder str = new StringBuilder();
            string colDelim = colJoin;
            string rowDelim = "";
            int count = Rows.Count;
            for (int row = 0; row < count; ++row)
            {
                str.Append(rowDelim + _Str(row, column1, "")
                    + colDelim + _Str(row, column2, ""));
                rowDelim = rowJoin;
            }
            return str.ToString();
        }
        public string _Collate(string column1, string column2, string column3, string colJoin, string rowJoin)
        {
            StringBuilder str = new StringBuilder();
            string colDelim = colJoin;
            string rowDelim = "";
            int count = Rows.Count;
            for (int row = 0; row < count; ++row)
            {
                str.Append(rowDelim + _Str(row, column1, "")
                    + colDelim + _Str(row, column2, "")
                    + colDelim + _Str(row, column3, "")
                    );
                rowDelim = rowJoin;
            }
            return str.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ColumnValues -->
        /// <summary>
        ///      Returns the values in a column
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public List<object> _ColumnValues(string column)
        {
            List<object> list = new List<object>();
            int count = Count;
            for (int row = 0; row < count; ++row)
            {
                list.Add(Rows[row][column]);
                LastRow = row;
            }
            LastColumn = column;
            return list;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Copy -->
        /// <summary>
        ///      Copies the entire RichDataTable including the current errors and xml if they exist
        /// </summary>
        /// <returns></returns>
        public RichDataTable Copy()
        {
            RichDataTable copy = new RichDataTable(_table.Copy());
            copy._errors     = _errors;
            copy._results    = _results;
            copy._lastRow    = _lastRow;
            copy._lastColumn = _lastColumn;
            copy._lastXml    = _lastXml;
            if (_lastXdoc != null)
                copy._lastXdoc = _lastXdoc._Copy();
            else copy._lastXdoc = null;
            return copy;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Count -->
        /// <summary>
        ///      Counts the number rows containing items matching the input pattern
        /// </summary>
        /// <param name="table"></param>
        /// <param name="column"></param>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int _Count(string column, string pattern)
        {
            int total = 0;


            if (_table != null && Columns.Contains(column))
            {
                int count = Count;
                for (int row = 0; row < count; ++row)
                    if (Regex.IsMatch(_Str(row, column, ""), pattern))
                        total += 1;
            }


            return total;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Count -->
        /// <summary>
        ///      Counts the number of matches in a string representation of the cell to the pattern
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int _Count(int row, string column, string pattern)
        {
            Regex regex = new Regex(pattern);
            string str = _Str(row, column, "");
            MatchCollection matches = regex.Matches(str);
            return matches.Count;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Date -->
        /// <summary>
        ///      Returns a Date from a cell
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="kind"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public Date _Date(int row, string column, DateTimeKind kind, Date defaultValue)
        {
            Date date;
            if (0 <= row && row < Count && _table.Columns.Contains(column))
            {
                object obj = _table.Rows[row][column];
                if (obj is SqlDateTime)
                    date = (new Date((SqlDateTime)obj, kind));
                else
                    date = (new Date(TreatAs.StrValue(obj, defaultValue.ToString()), kind));
            }
            else date = defaultValue;


            this._RecordLast(row, column);
            return date;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Date_sql -->
        /// <summary>
        ///      Returns a SqlDateTime which allows null dates
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue">may use SqlDateTime.Null</param>
        /// <returns></returns>
        public SqlDateTime _Date_sql(int row, string column, SqlDateTime defaultValue)
        {
            SqlDateTime date;
            if (this._Exists(row, column)) date = Date.SqlDate(_table.Rows[row][column], defaultValue);
            else date = defaultValue;
            this._RecordLast(row, column);
            return date;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Date_sql -->
        /// <summary>
        ///      Returns a SqlDateTime given a number of possibilities for column name
        /// </summary>
        /// <param name="row"></param>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <param name="d"></param>
        /// <param name="e">5th item</param>
        /// <param name="f"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public SqlDateTime _Date_sql(int row, string[] possibleColumn, SqlDateTime defaultValue)
        {
            SqlDateTime date = defaultValue;


            foreach (string column in possibleColumn)
            {
                date = this._Date_sql(row, column, defaultValue);
                if (date.ToString() != defaultValue.ToString())
                    break;
            }
            LastRow = row;
            return date;
        }
        public SqlDateTime _Date_sql(int row, string a, string b,                                                   SqlDateTime defaultValue) { string[] cols = { a, b };                 return this._Date_sql(row, cols, defaultValue); }
        public SqlDateTime _Date_sql(int row, string a, string b, string c,                                         SqlDateTime defaultValue) { string[] cols = { a, b, c };              return this._Date_sql(row, cols, defaultValue); }
        public SqlDateTime _Date_sql(int row, string a, string b, string c, string d,                               SqlDateTime defaultValue) { string[] cols = { a, b, c, d };           return this._Date_sql(row, cols, defaultValue); }
        public SqlDateTime _Date_sql(int row, string a, string b, string c, string d, string e,                     SqlDateTime defaultValue) { string[] cols = { a, b, c, d, e };        return this._Date_sql(row, cols, defaultValue); }
        public SqlDateTime _Date_sql(int row, string a, string b, string c, string d, string e, string f,           SqlDateTime defaultValue) { string[] cols = { a, b, c, d, e, f };     return this._Date_sql(row, cols, defaultValue); }
        public SqlDateTime _Date_sql(int row, string a, string b, string c, string d, string e, string f, string g, SqlDateTime defaultValue) { string[] cols = { a, b, c, d, e, f, g };  return this._Date_sql(row, cols, defaultValue); }


        // ----------------------------------------------------------------------------------------
        /// <!-- _DateTime -->
        /// <summary>
        ///      Returns a regular System.DateTime (which does not allow null dates so this can be a problem)
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public DateTime _DateTime(int row, string column, DateTime defaultValue)
        {
            DateTime date;
            if (this._Exists(row, column))
                date = Date.ClrDate(_table.Rows[row][column], defaultValue);
            else date = defaultValue;
            this._RecordLast(row, column);
            return date;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Dec -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public decimal _Dec(int row, string column, decimal defaultValue)
        {
            decimal num;
            if (this._Exists(row,column))
                num = TreatAs.DecValue(_table.Rows[row][column], defaultValue);
            else num = defaultValue;
            this._RecordLast(row, column);
            return num;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Distinct -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public List<object> _Distinct(string column)
        {
            Dictionary<string, int> hash = new Dictionary<string,int>();
            List<object> list = new List<object>();
            bool nullInList = false;
            for (int row = 0; row < Count; ++row)
            {
                object obj = this._Obj(row, column, null);
                if (obj == null && !nullInList)
                {
                    list.Add(null);
                    nullInList = true;
                }
                else
                {
                    string key = obj.ToString();
                    if (!hash.ContainsKey(key))
                    {
                        hash.Add(key, 1);
                        list.Add(obj);
                    }
                }
            }
            return list;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _RandomValue -->
        /// <summary>
        ///      Returns the number of a random row in string format
        /// </summary>
        /// <param name="r"></param>
        /// <param name="defaultString"></param>
        /// <returns></returns>
        public string _RandomValue(Random r, string defaultString)
        {
            return this._Str(r.Next(Count), 0, defaultString);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Endeme -->
        /// <summary>
        ///      Returns an endeme contained as single characters in a series of columns
        /// </summary>
        /// <param name="columns"></param>
        /// <param name="row"></param>
        /// <returns></returns>
        public Endeme _Endeme(int row, List<string> columns)
        {
            string str = "";
            Dictionary<char,int> positions = new Dictionary<char,int>(columns.Count);
            for (int i = 0; i < columns.Count; ++i)
                str += _Char(row, columns[i], 'X');
            return new Endeme(null, str);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExcelToTable -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="excelPath">path to the excel spreadsheet</param>
        /// <param name="worksheet">the worksheet to extract</param>
        /// <returns></returns>
        public static RichDataTable ExcelToTable(string excelPath, string worksheet)
        {
            // --------------------------------------------------------------------------
            //  Prepare command
            // --------------------------------------------------------------------------
            string sourceConstr = @"Provider=Microsoft.ACE.OLEDB.12.0;"
                + "Data Source='" + excelPath + "';"
                + "Extended Properties='Excel 12.0;HDR=No;IMEX=1'";
            OleDbConnection con = new OleDbConnection(sourceConstr);
            OleDbDataAdapter myCommand = new OleDbDataAdapter(" SELECT * FROM ["+worksheet+"$]", con);


            // --------------------------------------------------------------------------
            //  Get the data
            // --------------------------------------------------------------------------
            DataSet ds = new DataSet();
            myCommand.TableMappings.Add(worksheet, "dtExcel");
            myCommand.Fill(ds);
            RichDataTable data = new RichDataTable(ds.Tables[0]);


            return data;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Exists -->
        /// <summary>
        ///      Determines whether the specified (row X column) exists
        /// </summary>
        /// <param name="row">row number</param>
        /// <param name="column">column name</param>
        private bool _Exists(int row, string column) { return (0 <= row && row < Count && _table.Columns.Contains(column)       ); } /// <param name="row">row number</param><param name="col">column number</param>
        private bool _Exists(int row, int col)       { return (0 <= row && row < Count && 0 <= col && col < _table.Columns.Count); }


        // ----------------------------------------------------------------------------------------
        /// <!-- _FillFrom -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="dr"></param>
        /// <param name="disposeSourceReader"></param>
        /// <returns></returns>
        public SqlDataReader _FillFrom(SqlDataReader dr, bool disposeSourceReader)
        {
            DataReaderAdapter dra = new DataReaderAdapter();
            _table = new DataTable();
            dra.FillFromReader(_table, dr);
            if (disposeSourceReader) { dr.Dispose(); dr = null; }
            return dr;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Float -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public float _Float(int row, string column, float defaultValue)
        {
            float num;
            if (this._Exists(row, column))
                num = TreatAs.FloatValue(_table.Rows[row][column], defaultValue);
            else num = defaultValue;
            this._RecordLast(row, column);
            return num;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Guid -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="p"></param>
        /// <param name="p_3"></param>
        /// <returns></returns>
        public Guid _Guid(int row, string column, Guid defaultValue)
        {
            Guid guid;
            if (this._Exists(row, column))
                guid = TreatAs.GuidValue(_table.Rows[row][column], defaultValue);
            else guid = defaultValue;
            this._RecordLast(row, column);
            return guid;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Guids -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public Guid[] _Guids(string column)
        {
            int count = Count;
            Guid[] array = new Guid[count];
            for (int row = 0; row < count; ++row)
            {
                array[row] = this._Guid(row, column, Guid.Empty);
                LastRow = row;
            }
            LastColumn = column;
            return array;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Int -->
        /// <summary>
        ///      Returns an exception-free integer representation of the cell value
        /// </summary>
        /// <param name="row">cell row</param>
        /// <param name="column">cell column</param>
        /// <param name="defaultValue">value returned if the method can not figure it out</param>
        /// <returns></returns>
        public int _Int(int row, string column, int defaultValue)
        {
            int num;
            if (this._Exists(row, column))
                num = TreatAs.IntValue(_table.Rows[row][column], defaultValue);
            else num = defaultValue;
            this._RecordLast(row, column);
            return num;
        }
        public int _Int(int row, int col, int defaultValue)
        {
            if (0 <= col && col < Columns.Count)
                return _Int(row, Columns[col].ColumnName, defaultValue);
            string error = "Error - column index out of range in RichDataTable._Int(row, col): " + col;
            Throws.A(new ArgumentOutOfRangeException(error));
            _RecordLast(row, error);
            return defaultValue;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Max -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public DateTime _Max(string column, DateTime defaultValue)
        {
            DateTime max = Date.MinSqlValue.CLRFormat;
            if (this._Exists(0, column))
            {
                for (int row = 0; row < Count; ++row)
                    if (_DateTime(row, column, defaultValue) > max)
                        max = _DateTime(row, column, defaultValue);
            }
            else
                max = defaultValue;
            return max;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Min -->
        /// <summary>
        ///      Returns the minimum value of a column
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public int _Min(string column)
        {
            int count = Rows.Count;
            int minval = (int)Rows[0][column];
            for (int row = 0; row < count; ++row)
                minval = Math.Min(minval, (int)Rows[row][column]);
            return minval;
        }
        public DateTime _Min(string column, DateTime defaultValue)
        {
            DateTime min = DateTime.MaxValue;
            if (this._Exists(0, column))
            {
                for (int row = 0; row < Count; ++row)
                    if (_DateTime(row, column, defaultValue) < min)
                        min = _DateTime(row, column, defaultValue);
            }
            else
                min = defaultValue;
            return min;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _MinimizeXml -->
        /// <summary>
        ///      Removes all empty tags
        /// </summary>
        /// <remarks>
        ///      Move this to an XML namespace!
        /// </remarks>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        public string _MinimizeXml(int row, string column)
        {
            string small = Minimize(this._Str(row, column, ""));
            this._RecordLast(row, column);
            return small;
        }
        private string Minimize(string xml)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xml);
            Minimize(doc.FirstChild, 0);
            return doc.InnerXml;
        }
        private XmlNode Minimize(XmlNode node, int level)
        {
            for (int i =  node.ChildNodes.Count - 1; i>= 0; --i)
            {
                XmlNode child =  node.ChildNodes[i];
                if (child.InnerText == "")
                    node.RemoveChild(child);
                else if (child.InnerXml != child.InnerText)
                    Minimize(child, level + 1);
            }
            return node;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _NewColumnList -->
        /// <summary>
        ///      Creates a new column list from the object minus the specified column
        /// </summary>
        /// <param name="columnToMakeIntoTables"></param>
        /// <returns></returns>
        private DataColumn[] _NewColumnList(string columnToLeaveOut)
        {
            int count = _table.Columns.Count;
            DataColumn[] newColumns;


            if (_table.Columns.Contains(columnToLeaveOut))
            {
                // ---------------------------------------------------------------------90
                //  Set up the initial array structures
                // ---------------------------------------------------------------------90
                DataColumn[] oldColumns = new DataColumn[count];
                newColumns = new DataColumn[count - 1];
                _table.Columns.CopyTo(oldColumns, 0);


                // ---------------------------------------------------------------------90
                //  Add all the columns except the specified one
                // ---------------------------------------------------------------------90
                int x = 0;
                for (int col = 0; col < count; ++col, ++x)
                {
                    if (oldColumns[col].ColumnName == columnToLeaveOut) --x;
                    else newColumns[x] = oldColumns[col];
                }
            }
            else
            {
                // ---------------------------------------------------------------------90
                //  If the column to leave out was not in the list, just copy the list
                // ---------------------------------------------------------------------90
                newColumns = new DataColumn[count];
                _table.Columns.CopyTo(newColumns, 0);
            }


            return newColumns;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _NewTable -->
        /// <summary>
        ///      Creates a new table out of a column array and a name
        /// </summary>
        /// <param name="tableName"></param>
        /// <param name="newColumns"></param>
        /// <returns></returns>
        private static DataTable _NewTable(string tableName, DataColumn[] newColumns)
        {
            DataTable table = new DataTable();
            foreach (DataColumn col in newColumns)
                table.Columns.Add(col.ColumnName, col.DataType);
            table.TableName = tableName;
            return table;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Obj -->
        /// <summary>
        ///      Returns an exception-free object representation of the cell value
        /// </summary>
        /// <param name="row">cell row</param>
        /// <param name="column">cell column</param>
        /// <param name="defaultValue">value returned if the cell is null or does not exist</param>
        /// <returns></returns>
        public object _Obj(int row, string column, object defaultValue)
        {
            object obj;
            if (this._Exists(row, column))
                obj = _table.Rows[row][column];
            else obj = defaultValue;
            this._RecordLast(row, column);
            return obj;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _RecordLast -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        private void _RecordLast(int row, string column)
        {
            LastRow    = row;
            LastColumn = column;
        }
        private void _RecordLast(int row, string column, RichXmlDocument xdoc, string xml)
        {
            LastRow    = row;
            LastColumn = column;
            LastXml    = xml;
            LastXdoc   = xdoc;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ChangeType -->
        /// <summary>
        ///      Changes the type of a column and migrates the data
        /// </summary>
        /// <param name="column"></param>
        /// <param name="type"></param>
        /// <param name="defaultValue"></param>
        public void _ChangeType(string column, Type type, double defaultValue)
        {
            if (Columns.Contains(column))
            {
                _Rename(column, "Jesus");
                _AddColumn(column, type);
                for (int row = 0; row < Count; ++row)
                    this.Rows[row][column] = _Float(row, "Jesus", (float)defaultValue);
                Columns.Remove("Jesus");
            }
        }
        public void _ChangeType(string column, Type type, int defaultValue)
        {
            if (Columns.Contains(column))
            {
                _Rename(column, "Jesus");
                _AddColumn(column, type);
                for (int row = 0; row < Count; ++row)
                    this.Rows[row][column] = _Int(row, "Jesus", defaultValue);
                Columns.Remove("Jesus");
            }
        }
        public void _ChangeType(string column, Type type, string defaultValue)
        {
            if (this.Columns.Contains(column))
            {
                _Rename(column, "Jesus");
                _AddColumn(column, type);
                for (int row = 0; row < Count; ++row)
                    this.Rows[row][column] = _Str(row, "Jesus", defaultValue);
                Columns.Remove("Jesus");
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Rename -->
        /// <summary>
        ///      Renames a column if the column exists
        /// </summary>
        /// <param name="col"></param>
        /// <param name="column"></param>
        public void _Rename(int col, string column)
        {
            if (col < this.Columns.Count)
                Columns[col].ColumnName = column;
        }
        public void _Rename(string columnFrom, string columnTo)
        {
            if (this.Columns.Contains(columnFrom))
                Columns[columnFrom].ColumnName = columnTo;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Select -->
        /// <summary>
        ///      Returns a subtable from a table using a filter
        /// </summary>
        /// <param name="from"></param>
        /// <param name="filter"></param>
        /// <returns></returns>
        public RichDataTable _Select(string filter)
        {
            DataRow[] rows = this.Select(filter);
            RichDataTable subTable = this.Clone();
            foreach (DataRow oldRow in rows)
                subTable.Table.ImportRow(oldRow);
            return subTable;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Sets the value of a row/column to an xml
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="xml"></param>
        public void _SetColumn(int row, string column, SqlXml xml)
        {
            this.Table.Rows[row][column] = xml;
            this._RecordLast(row, column);
        }
        /// <summary>
        ///      Sets the value of a row/column to a string
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="str"></param>
        public void _SetColumn(int row, string column, string str)
        {
            this.Table.Rows[row][column] = str;
            this._RecordLast(row, column);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Sets the value of a row/column based on another column
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="columnFrom"></param>
        /// <param name="pattern"></param>
        /// <param name="trueValue"></param>
        /// <param name="falseValue"></param>
        public void _SetColumn(int row, string columnTo
            , string columnFrom, string pattern
            , string trueValue, string falseValue)
        {
            string str = this._Str(row, columnFrom, "");
            if (Regex.IsMatch(str, pattern))
                this.Table.Rows[row][columnTo] = trueValue;
            else
                this.Table.Rows[row][columnTo] = falseValue;
            LastRow = row;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Concatenates the values of two columns into a third
        /// </summary>
        /// <param name="columnTo"></param>
        /// <param name="column1"></param>
        /// <param name="delim"></param>
        /// <param name="column2"></param>
        public void _SetColumn(string columnTo, string column1, string delim, string column2)
        {
            if (this._Exists(0, columnTo) && this._Exists(0, column1) && this._Exists(0, column2))
            {
                for (int row = 0; row < Count; ++row)
                    this.Table.Rows[row][columnTo] = _Str(row, column1, "") + delim + _Str(row, column2, ""); 
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Sets one column to another (if the 'to column' is a string type)
        /// </summary>
        /// <param name="columnTo"></param>
        /// <param name="columnFrom"></param>
        public void _SetColumn(string columnTo, string columnFrom, string defaultvalue)
        {
            if (this._Exists(0, columnTo) && this._Exists(0, columnFrom))
            {
                for (int row = 0; row < Count; ++row)
                    this.Table.Rows[row][columnTo]
                        = _Str(row, columnFrom, defaultvalue);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Sets one column to another (if the 'to column' is a string type)
        /// </summary>
        /// <param name="columnTo"></param>
        /// <param name="columnFrom"></param>
        public void _SetColumn(string columnTo, string columnFrom, float defaultvalue)
        {
            if (this._Exists(0, columnTo) && this._Exists(0, columnFrom))
            {
                for (int row = 0; row < Count; ++row)
                    this.Table.Rows[row][columnTo]
                        = _Float(row, columnFrom, defaultvalue);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetColumn -->
        /// <summary>
        ///      Sets all the values of a column to a particular literal
        /// </summary>
        /// <param name="columnTo"></param>
        /// <param name="defaultValue"></param>
        /// <param name="fleet"></param>
        public void _SetColumn(string columnTo, object value, string defaultValue)
        {
            if (this._Exists(0, columnTo))
            {
                string str = TreatAs.StrValue(value, defaultValue);
                for (int row = 0; row < Count; ++row)
                    this.Table.Rows[row][columnTo] = str; 
            }
        }


        // ----------------------------------------------------------------------------------------
        //  _SetColumn
        // ----------------------------------------------------------------------------------------
        /// <summary>Sets all of the values in a column to a numeric literal</summary>
        public void _SetColumn(string column, double value) { if (this._Exists(0, column)) for (int row = 0; row < Count; ++row) _table.Rows[row][column] = value; } /// <summary>Sets all of the values in a column to a string literal</summary>
        public void _SetColumn(string column, string value) { if (this._Exists(0, column)) for (int row = 0; row < Count; ++row) _table.Rows[row][column] = value; } /// <summary>Sets all of the values in a column to an int literal</summary>
        public void _SetColumn(string column, int    value) { if (this._Exists(0, column)) for (int row = 0; row < Count; ++row) _table.Rows[row][column] = value; } /// <summary>Sets all of the values in a column to a float literal</summary>
        public void _SetColumn(string column, float  value) { if (this._Exists(0, column)) for (int row = 0; row < Count; ++row) _table.Rows[row][column] = value; }


        // ----------------------------------------------------------------------------------------
        /// <!-- _AllTheSame -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public bool _AllTheSame(string column)
        {
            bool theSame = true;
            if (!_table.Columns.Contains(column)) return false;
            else
            {
                if (Count == 0) return true;
                else
                {
                    string value;
                    string firstValue = _Str(0, column, "");
                    for (int row = 0; row < Count && theSame; ++row)
                    {
                        value = _Str(row, column, "");
                        if (value != firstValue) theSame = false;
                    }
                    return theSame;
                }
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Sort -->
        /// <summary>
        ///      Returns the RichDataTable after being sorted by passing it through a DataView.
        /// </summary>
        // ----------------------------------------------------------------------------------------
        public void Sort(string orderByClause)
        {
            // -------------------------------------------------------------------------90
            //  Load the DataTable into a view
            // -------------------------------------------------------------------------90
            DataView dv = new DataView(_table);


            // -------------------------------------------------------------------------90
            //  Prepare for the sort
            // -------------------------------------------------------------------------90
            orderByClause = Regex.Replace(orderByClause, "^ *order +by +", "", RegexOptions.IgnoreCase);
            orderByClause = Regex.Replace(orderByClause, "^ *by +", "", RegexOptions.IgnoreCase);
            try
            {
                dv.Sort = orderByClause;
                dv.ApplyDefaultSort = true;


                // ---------------------------------------------------------------------90
                //  Load the view back into the DataTable
                // ---------------------------------------------------------------------90
                RichDataTable rt = new RichDataTable(dv.ToTable());
                _table.Clear();
                _table.Merge(rt._table);
            }
            catch (Exception ex) { _errors = ex.Message; }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Split -->
        /// <summary>
        ///      Splits a table into multiple tables based on the values in a column,
        ///      This is part an effort to deal with the famous SQL/C# impedence mismatch
        /// </summary>
        /// <param name="column"></param>
        public List<RichDataTable> _Split(string column)
        {
            List<object> keys = this._Distinct(column);
            Dictionary<object,RichDataTable> tables = new Dictionary<object,RichDataTable>(keys.Count);
            foreach (object obj in keys)
                tables[obj] = this.Clone();


            for (int row = 0; row < Count; ++row)
            {
                object obj = this._Obj(row, column, null);
                tables[obj].Add(this, row);
            }


            List<RichDataTable> table = new List<RichDataTable>(keys.Count);
            for (int i = 0; i < keys.Count; ++i)
            {
                object obj = keys[i];
                table.Add(tables[obj]);
            }


            return table;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Str -->
        /// <summary>
        ///      Converts a cell value to a string, using the default value when null or nonexistent,
        ///      This method never blows up so exceptions need to be thrown above this level in the code
        /// </summary>
        /// <param name="row">cell row</param>
        /// <param name="column">cell column</param>
        /// <param name="defaultValue">
        ///      value returned if the cell is nonexistent or its value is null
        /// </param>
        /// <returns></returns>
        public string _Str(int row, string column, string defaultValue)
        {
            string str;
            if (this._Exists(row, column))
                str = TreatAs.StrValue(_table.Rows[row][column], defaultValue);
            else str = defaultValue;
            this._RecordLast(row, column);
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Str -->
        /// <summary>
        ///      Converts a cell value to a string, using the default value when null or nonexistent,
        ///      This method never blows up so exceptions need to be thrown above this level in the code
        /// </summary>
        /// <param name="row"></param>
        /// <param name="col"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public string _Str(int row, int col, string defaultValue)
        {
            if (0 <= row && row < Count && 0 <= col && col < _table.Columns.Count)
                return TreatAs.StrValue(_table.Rows[row][col], defaultValue);
            _RecordLast(row, "Error - row or column number out of range: "+ row + "x" + col);
            return defaultValue;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Str -->
        /// <summary>
        ///      Returns the first column's value that is not blank
        /// </summary>
        /// <param name="row"></param>
        /// <param name="possibleColumn">array of possible column names</param>
        /// <param name="defaultValue">default value to look for something else than</param>
        /// <returns></returns>
        private string _Str(int row, string[] possibleColumn)
        {
            string str = "";
            int count = possibleColumn.Length;
            for (int i = 0; i < count && str == ""; ++i)
                str = this._Str(row, possibleColumn[i], "");
            return str;
        }
        public string _Str(int row, string a, string b, string c)                               { string[] cols = { a, b, c };           return this._Str(row, cols); }
        public string _Str(int row, string a, string b, string c, string d)                     { string[] cols = { a, b, c, d };        return this._Str(row, cols); }
        public string _Str(int row, string a, string b, string c, string d, string e)           { string[] cols = { a, b, c, d, e };     return this._Str(row, cols); }
        public string _Str(int row, string a, string b, string c, string d, string e, string f) { string[] cols = { a, b, c, d, e, f };  return this._Str(row, cols); }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Strings -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column"></param>
        /// <returns></returns>
        public string[] _Strings(string column)
        {
            int count = Count;
            string[] array = new string[count];
            for (int row = 0; row < count; ++row)
                array[row] = this._Str(row, column, "");
            return array;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Strings -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="column1"></param>
        /// <param name="join"></param>
        /// <param name="column2"></param>
        /// <returns></returns>
        public string[] _Strings(string column1, string literal, string column2)
        {
            int count = Count;
            string[] array = new string[count];
            for (int i = 0; i < count; ++i)
                array[i] = this._Str(i, column1, "") + literal + this._Str(i, column2, "");
            return array;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Sum -->
        /// <summary>
        ///      Returns the sum of the values in a column
        /// </summary>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public float _Sum(string column, float defaultValue)
        {
            float sum = 0.0F;
            if (_table != null && this._Exists(0, column))
                for (int row = 0; row < Count; ++row) sum += _Float(row, column, 0.0F);
            else sum = defaultValue;
            return sum;
        }
        public float _Sum(string column, int defaultValue)
        {
            int sum = 0;
            if (_table != null && this._Exists(0, column))
                for (int row = 0; row < Count; ++row) sum += _Int(row, column, 0);
            else sum = defaultValue;
            return sum;
        }
        public double _Sum(string column, double defaultValue)
        {
            double sum = 0.0;
            if (_table != null && this._Exists(0, column))
                for (int row = 0; row < Count; ++row) sum += _Float(row, column, 0.0F);
            else sum = defaultValue;
            return sum;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _TallyColumnValues -->
        /// <summary>
        ///      Prodcues a tally of the values in a column
        /// </summary>
        /// <param name="biz"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        public Tally _TallyColumnValues(string column, bool includeBlanks)
        {
            Tally tally = new Tally();
            for (int row = 0; row < Count; ++row)
            {
                string value = _Str(row, column, "");
                tally.Incr(value, includeBlanks);
            }
            return tally;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ToDataSet -->
        /// <summary>
        ///      Converts a data table into a dataset using a column for the DataSet table names
        /// </summary>
        /// <param name="columnToMakeIntoTables">The column containing the DataSet table names</param>
        /// <returns></returns>
        public DataSet _ToDataSet(string columnToMakeIntoTables)
        {
            DataSet ds = new DataSet();


            if (_table.Columns.Contains(columnToMakeIntoTables))
            {
                int count = Count;
                string tableName;
                DataTable dt;
                int newRow;


                // ---------------------------------------------------------------------90
                //  Construct the new table columns list
                // ---------------------------------------------------------------------90
                DataColumn[] newColumns = this._NewColumnList(columnToMakeIntoTables);


                for (int row = 0; row < count; ++row)
                {
                    // -----------------------------------------------------------------90
                    //  Educe a table reference
                    // -----------------------------------------------------------------90
                    tableName = this._Str(row, columnToMakeIntoTables, "default_table");
                    if (!ds.Tables.Contains(tableName))
                        ds.Tables.Add(_NewTable(tableName, newColumns));
                    dt = ds.Tables[tableName];


                    // -----------------------------------------------------------------90
                    //  Create a new row
                    // -----------------------------------------------------------------90
                    newRow = dt.Rows.Count;
                    dt.Rows.Add(dt.NewRow());
                    foreach (DataColumn col in newColumns)
                        dt.Rows[newRow][col.ColumnName]
                            = _table.Rows[row][col.ColumnName];
                }
            }


            return ds;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return TreatAs.StrValue(TableName, "") + "(" + Count.ToString() + ")";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _TrimColumn -->
        /// <summary>
        ///      Trims leading and trailing spaces from every value in a string column
        /// </summary>
        /// <param name="column"></param>
        public void _TrimColumn(string column, string defaultValue)
        {
            if (this._Exists(0, column))
            {
                for (int row = 0; row < Count; ++row)
                   _table.Rows[row][column] = _Str(row, column, defaultValue).Trim(); 
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Xml -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public RichXmlDocument _Xml(int row, string column)
        {
            RichXmlDocument xdoc;
            string xml = "";
            if (this._Exists(row, column))
            {
                xml = this._Str(row, column, "");
                xdoc = new RichXmlDocument(xml);
            }
            else xdoc = new RichXmlDocument();
            this._RecordLast(row, column, xdoc, xml);
            return xdoc;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _XmlDate -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="nameSpace"></param>
        /// <param name="xpath"></param>
        /// <returns></returns>
        public Date _XmlDate(int row, string column, string nameSpace, string xpath)
        {
            string text = this._XmlStr(row, column, nameSpace, xpath);
            Date date = new Date(text);
            this._RecordLast(row, column);
            return date;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _XmlNodes -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <param name="nameSpace"></param>
        /// <param name="xpath"></param>
        /// <returns></returns>
        public XmlNodeList _XmlNodes(int row, string column, string nameSpace, string xpath)
        {
            RichXmlDocument xdoc = null;
            XmlNodeList node = null;
            //string text = "";
            string xml = "";


            if (this._Exists(row, column))
            {
                //  Get xml
                xml = this._Str(row, column, "");
                if (xml != this.LastXml || string.IsNullOrEmpty(_lastXml))
                    xdoc = new RichXmlDocument(xml); // WARNING: inefficient
                else xdoc = this.LastXdoc;


                if (xdoc != null) node = xdoc._GetNodes(nameSpace, xpath);
            }


            this._RecordLast(row, column, xdoc, xml);
            return node;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _XmlStr -->
        /// <summary>
        ///      Returns a field in an xml in a table row specified by an xpath
        /// </summary>
        /// <remarks>
        /// WARNING: inefficient,
        ///      Inefficient if you want more than one field because it converts the xml string
        ///      to an XmlDocument for each field, For more than one field use
        ///          _GetNode([namespace], [xpath]).Innertext   in RichXmlDocument
        ///      instead
        /// UPDATE: fixed the inefficiency problem
        /// </remarks>
        /// <param name="row">row number:  0 to n-1</param>
        /// <param name="column">The column the xml document is in</param>
        /// <param name="nameSpace">The general namespace</param>
        /// <param name="xpath">An xpath to the field</param>
        /// <returns>returns an xml field, defaults to ""</returns>
        public string _XmlStr(int row, string column, string nameSpace, string xpath)
        {
            RichXmlDocument xdoc = null;
            XmlNode node = null;
            string text = "";
            string xml = "";


            if (this._Exists(row, column))
            {
                // ---------------------------------------------------------------------90
                //  Get xml, This WAS inefficient, now it caches the xml document
                // ---------------------------------------------------------------------90
                xml = this._Str(row, column, "");
                if (xml != this.LastXml || string.IsNullOrEmpty(_lastXml))
                    xdoc = new RichXmlDocument(xml);
                else xdoc = this.LastXdoc;


                if (xdoc != null) node = xdoc._GetNode(nameSpace, xpath);
                if (node != null) text = node.InnerText;
            }


            this._RecordLast(row, column, xdoc, xml);
            return text;
        }


        public string _ColumnList()
        {
            string str = "";
            string delim = "";
            foreach (DataColumn column in _table.Columns)
            { str += delim + column.ColumnName; delim = ", "; }
            return str;
        }


        #endregion methods and properties


        

        // ----------------------------------------------------------------------------------------
        /// <!-- PrettyColumnList -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="rdtLaneAnalysis"></param>
        /// <returns></returns>
        public string PrettyColumnList(RichDataTable rdtLaneAnalysis)
        {
            DataColumnCollection column = rdtLaneAnalysis.Columns;
            string str = "";
            foreach (DataColumn col in column)
            {
                string name = col.ColumnName;
                Type type = col.DataType;
                str += "\r\n" + name + "  ---  " + type.Name;
            }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- LinqToDataTable -->
        /// <summary>
        ///      this is very fragile - this linq has to be exactly right and any changes to the
        ///      datatable will break it
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="varlist"></param>
        /// <returns></returns>
        public DataTable LinqToDataTable<T>(IEnumerable<T> varlist)
        {
            // --------------------------------------------------------------------------
            //  Prep
            // --------------------------------------------------------------------------
            DataTable output = new DataTable();
            PropertyInfo[] property = null;  // column names 
            if (varlist == null) return output;


            foreach (T rec in varlist)
            {
                // -----------
                //  Use reflection to get property names, to create table, Only first time, others will follow 
                if (property == null)
                {
                    property = ((Type)rec.GetType()).GetProperties();
                    foreach (PropertyInfo pi in property)
                    {
                        Type colType = pi.PropertyType;
                        if ((colType.IsGenericType) && (colType.GetGenericTypeDefinition() == typeof(Nullable<>)))
                            colType = colType.GetGenericArguments()[0];
                        output.Columns.Add(new DataColumn(pi.Name, colType));
                    }
                }

                DataRow dr = output.NewRow();

                foreach (PropertyInfo pi in property)
                {
                    dr[pi.Name] = pi.GetValue(rec, null) == null ? DBNull.Value : pi.GetValue
                    (rec, null);
                }

                output.Rows.Add(dr);
            }
            return output;
        }

    }
}
