// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using InformationLib.Endemes;                     // for Endeme(6)
using InformationLib.SoftData;                    // for Endeme(6)
using InformationLib.Strings;                     // for IResults(1)
using System;                                     // for TimeSpan(2), Guid(1), EventArgs(1)
using System.Collections.Generic;                 // for List
using System.ComponentModel;                      // for IContainer
using System.Configuration;                       // for ConfigurationManager
using System.Data;                                // for IDataParameterCollection
using System.Data.Sql;                            // for SqlNotificationRequest
using System.Data.SqlClient;                      // for SqlCOmmand, SqlConnection
using System.Data.SqlTypes;                       // for SqlTypeException
using System.Text.RegularExpressions;             // for Regex
using System.Windows.Forms;
using System.Xml;                                 // for XmlReader

namespace InformationLib.DataAccess
{
    // --------------------------------------------------------------------------------------------
    /// <!-- RichSqlCommand -->
    /// <summary>
    ///      The RichSqlCommand class is a pseudo-derived class of the SqlCommand class with a few
    ///      extra methods added
    /// </summary>
    public class RichSqlCommand
    {

        #region members and constructors


        // ----------------------------------------------------------------------------------------
        /// <!-- _Base -->
        /// <summary>
        ///      This object is from the class that I am enhancing, the SqlCommand class
        /// </summary>
        private SqlCommand _command;


        // ----------------------------------------------------------------------------------------
        //  private message return members
        // ----------------------------------------------------------------------------------------
        private string       _errors;        public string       Errors           { get { return _errors;        } set { _errors = value; } }
        private string       _results;       public string       Results          { get { return _results;       } }
        private int          _rowsAffected;  public int          _RowsAffected    { get { return _rowsAffected;  } }
        private int          _spReturns;     public int          _Returns         { get { return _spReturns;     } }
        private SqlParameter _returnValue;   public SqlParameter _ReturnParameter { get { return _returnValue;   } }
        private List<string> _inputParam;   // just for functions
        private TimeSpan     _operationTime; public TimeSpan     OperationTime    { get { return _operationTime; } }
        private Endeme       _exceptionAction;     public Endeme       ExceptionActions       { get { return _exceptionAction;     } }


        private bool _statementCompleted; public bool _StatementCompleted
        { get { return _statementCompleted; } set { _statementCompleted = value; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- RichSqlCommand constructors -->
        /// <summary>
        ///      Constructors which initialize variables, message handling, connections, commands
        /// </summary>
        public RichSqlCommand()
        {
            _InitVariables();
            _InitMessages();
            _exceptionAction = "SNL";
        }
        //public RichSqlCommand(Endeme exceptionAction, string cmdText)
        //{
        //    _InitVariables();
        //    _command.CommandText = cmdText;
        //    _InitMessages();
        //    _exceptionAction = exceptionAction;
        //}


        // ----------------------------------------------------------------------------------------
        /// <!-- RichSqlCommand constructor -->
        /// <summary>
        ///      Constructs a RichSqlCommand object to run a SQL function
        /// </summary>
        /// <param name="conn"></param>
        /// <param name="sqlFunction"></param>
        /// <param name="inputParam"></param>
        //public RichSqlCommand(string conn, string sqlFunction, string inputParam
        //    , EndemeSet actions, Endeme exceptionAction)
        //{
        //    _InitVariables();
        //    _inputParam.Add(inputParam);
        //    _command.CommandText = "SELECT "+sqlFunction+"("+inputParam+")";
        //    _command.Connection = ConnectSource.Connection(conn);
        //    _InitMessages();
        //    _command.CommandType = CommandType.Text;
        //    _exceptionAction = exceptionAction;
        //}
        //public RichSqlCommand(string conn, string sqlFunction, string inputParamA, string inputParamB
        //    , Endeme exceptionAction)
        //{
        //    _InitVariables();
        //    _inputParam.Add(inputParamA);
        //    _inputParam.Add(inputParamB);
        //    _command.CommandText = "SELECT "+sqlFunction+"("+inputParamA+", "+inputParamB+")";
        //    _command.Connection = ConnectSource.Connection(conn);
        //    _InitMessages();
        //    _command.CommandType = CommandType.Text;
        //    _exceptionAction = exceptionAction;
        //}


        // ----------------------------------------------------------------------------------------
        /// <!-- RichSqlCommand constructor -->
        /// <summary>
        ///      Instantiates a RichSqlCommand object with either a connection string
        ///      or a connection name in an app.config file
        /// </summary>
        /// <param name="cmdText"></param>
        /// <param name="commandType"></param>
        /// <param name="conn"></param>
        //public RichSqlCommand(CommandType type, string conn, Endeme exceptionAction, string cmdText)
        //{ Init(type, conn, exceptionAction, cmdText); }
        public RichSqlCommand(CommandType type, string conn, string cmdText, EndemeSet actions, Endeme exceptionAction)
        { Init(type, conn, exceptionAction, cmdText); }
        private void Init(CommandType type, string conn, Endeme exceptionAction, string cmdText)
        {
            if (type == CommandType.Text)
                ConsiderMakingThisAStoredProcedure();


            _InitVariables();
            _command.CommandText = cmdText;
            _command.Connection = ConnectSource.Connection(conn);
            _InitMessages();
            _command.CommandType = type;
            _exceptionAction = exceptionAction;
        }
        private void ConsiderMakingThisAStoredProcedure()
        { }


        #endregion members and constructors

        #region pass-through methods


        // ----------------------------------------------------------------------------------------
        //  Pseudo-inherited properties
        // ----------------------------------------------------------------------------------------
        public IDataParameterCollection Parameters { get { return _command.Parameters;             } }
        public SqlNotificationRequest Notification { get { return _command.Notification;           } set { _command.Notification = value; } }
        public UpdateRowSource UpdatedRowSource    { get { return _command.UpdatedRowSource;       } set { _command.UpdatedRowSource = value; } }
        public bool NotificationAutoEnlist         { get { return _command.NotificationAutoEnlist; } set { _command.NotificationAutoEnlist = value; } }


        public string          CommandText    { get { return _command.CommandText;    } set { _command.CommandText      = value; } }
        public int             CommandTimeout { get { return _command.CommandTimeout; } set { _command.CommandTimeout   = value; } }
        public CommandType     CommandType    { get { return _command.CommandType;    } set { _command.CommandType      = value; } }
        public SqlConnection   Connection     { get { return _command.Connection;     } set { _command.Connection       = value; } }
        public IContainer      Container      { get { return _command.Container;      } }
        public ISite           Site           { get { return _command.Site;           } set { _command.Site             = value; } }
        public SqlTransaction  Transaction    { get { return _command.Transaction;    } set { _command.Transaction      = value; } }


        // ----------------------------------------------------------------------------------------
        //  Pseudo-inherited methods - the Execute commands need to be written like _ExecuteNonQuery is
        // ----------------------------------------------------------------------------------------
        public void            Cancel()            {        _command.Cancel();           }
        public void            Dispose()           {        _command.Dispose();          }
        public void            Prepare()           {        _command.Prepare();          }
        public SqlCommand      _Clone()            { return _command.Clone();            } // TODO: rewrite to return RichSqlCommand
        public XmlReader       _ExecuteXmlReader() { return _command.ExecuteXmlReader(); } // TODO: rewrite like _ExecuteNonQuery
        public override string ToString()          { return _command.ToString();         }


        public object  GetLifetimeService()        { return _command.GetLifetimeService();        }
        public object  InitializeLifetimeService() { return _command.InitializeLifetimeService(); }
        public void    ResetCommandTimeout()       { _command.ResetCommandTimeout();              }


        #endregion pass-through methods

        #region new and rewritten methods


        // ----------------------------------------------------------------------------------------
        /// <!-- _AddParameter -->
        /// <summary>
        ///      Adds a parameter of the specified name, type and value
        /// </summary>
        /// <param name="paramName"></param>
        /// <param name="paramType"></param>
        /// <param name="paramValue"></param>
        public void _AddParameter(string paramName, DbType type, object value)
        {
            SqlParameter sp = new SqlParameter(paramName, type);
            sp.Value = value;
            sp.Direction = ParameterDirection.Input;
            _command.Parameters.Add(sp);
        }
        public void _AddParameter(string paramName, SqlDbType type, object value)
        {
            SqlParameter sp = new SqlParameter(paramName, type);
            sp.Value = value;
            sp.Direction = ParameterDirection.Input;
            _command.Parameters.Add(sp);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _AddParameter -->
        /// <summary>
        ///      Adds a parameter of the specified name, type, value and direction
        /// </summary>
        /// <param name="paramName"></param>
        /// <param name="paramType"></param>
        /// <param name="paramValue"></param>
        /// <param name="dir"></param>
        public void _AddParameter(string paramName, DbType type, object value, ParameterDirection dir)
        {
            SqlParameter sp = new SqlParameter(paramName, type);
            sp.Value = value;
            sp.Direction = dir;
            _command.Parameters.Add(sp);
        }
        public void _AddParameter(string paramName, SqlDbType type, object value, ParameterDirection dir)
        {
            SqlParameter sp = new SqlParameter(paramName, type);
            sp.Value = value;
            sp.Direction = dir;
            _command.Parameters.Add(sp);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Close -->
        /// <summary>
        ///      Close the connection if it was opened by this method - if it was closed
        /// </summary>
        /// <param name="closeConnection">closes connection if true</param>
        public void _Close(bool closeConnection)
        {
            if (closeConnection)
                Connection.Close();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _ConnectWith -->
        /// <summary>
        ///      Creates connection using connection name
        /// </summary>
        /// <param name="connectionName"></param>
        public void _ConnectWith(string connectionName)
        {
            _command.Connection.ConnectionString = ConfigurationManager
                .ConnectionStrings[connectionName].ConnectionString;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Date -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns></returns>
        public Date _Date(string parameter)
        {
            SqlParameter param = (SqlParameter)Parameters[parameter];
            Date date;
            if (param == null || param.Value == null)  date = Date.Null;
            else date = new Date(param.Value.ToString());
            return date;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteFunction -->
        /// <summary>
        ///      Executes a single valued prepared function
        /// </summary>
        /// <returns></returns>
        public object ExecuteFunction(SqlDbType inputType, object inputValue, object defaultValue)
        {
            // -------------------------------------------------------------------------90
            //  Set the parameter value
            // -------------------------------------------------------------------------90
            _RemoveParametersExcept("@Return_Value");
            _AddParameter(_inputParam[0], inputType, inputValue);


            // -------------------------------------------------------------------------90
            //  Run the function
            // -------------------------------------------------------------------------90
            bool ifOpenedhere = _Open();
            _StatementCompleted = false;
            object value;
            SqlDataReader rdr = null;
            try
            {
                // ---------------------------------------------------------------------90
                //  Run the function
                // ---------------------------------------------------------------------90
                rdr = this.ExecuteReader();
                if (rdr.HasRows)
                    { rdr.Read();  value = rdr[0]; }
                else value = defaultValue;
            }
            catch
            {
                Throws.A(new DataException("data reader failed on function "+_command.CommandText));
                value = defaultValue;
            }
            finally { if (rdr != null)  rdr.Dispose(); }
            _Close(ifOpenedhere);


            // -------------------------------------------------------------------------90
            //  Process errors detected by events
            // -------------------------------------------------------------------------90
            if (!string.IsNullOrEmpty(_errors))
                Throws.A(new DataException(_errors));
            if (!_statementCompleted)
            {
                string error = "statement '" + _command.CommandText + "' did not complete";
                _errors += "\r\n" + error;
                Throws.A(new SqlTypeException(error));
            }


            return value;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteFunction -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="typeA"></param>
        /// <param name="valueA"></param>
        /// <param name="typeB"></param>
        /// <param name="valueB"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public object ExecuteFunction(SqlDbType typeA, string valueA, SqlDbType typeB, string valueB
            , object defaultValue)
        {
            // -------------------------------------------------------------------------90
            //  Set the parameter value
            // -------------------------------------------------------------------------90
            _RemoveParametersExcept("@Return_Value");
            _AddParameter(_inputParam[0], typeA, valueA);
            _AddParameter(_inputParam[1], typeB, valueB);


            // -------------------------------------------------------------------------90
            //  Run the function
            // -------------------------------------------------------------------------90
            bool ifOpenedhere = _Open();
            _StatementCompleted = false;
            object value;
            SqlDataReader rdr = null;
            try
            {
                // ---------------------------------------------------------------------90
                //  Run the function
                // ---------------------------------------------------------------------90
                rdr = this.ExecuteReader();
                if (rdr.HasRows)
                    { rdr.Read();  value = rdr[0]; }
                else value = defaultValue;
            }
            catch
            {
                Throws.A(new DataException("data reader failed on function "+_command.CommandText));
                value = defaultValue;
            }
            finally { if (rdr != null)  rdr.Dispose(); }
            _Close(ifOpenedhere);


            // -------------------------------------------------------------------------90
            //  Process errors detected by events
            // -------------------------------------------------------------------------90
            if (!string.IsNullOrEmpty(_errors))
                Throws.A(new DataException(_errors), Throws.Actions, _exceptionAction);
            if (!_statementCompleted)
            {
                string error = "statement '" + _command.CommandText + "' did not complete";
                _errors += "\r\n" + error;
                Throws.A(new SqlTypeException(error), Throws.Actions, _exceptionAction);
            }


            return value;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteNonQuery_simple -->
        /// <summary>
        ///      Executes a Transact-SQL statement against the connection and
        ///      returns the number of rows affected
        /// </summary>
        /// <returns>the number of rows affected</returns>
        private int ExecuteNonQuery_simple()
        {
            DateTime stamp1 = DateTime.Now;
            _errors         = "";
            _statementCompleted = false;
            _rowsAffected   = _command.ExecuteNonQuery();


            _spReturns      = TreatAs.IntValue(_returnValue.Value, -2);
            DateTime stamp2 = DateTime.Now;
            _operationTime  = stamp2 - stamp1;
            return _rowsAffected;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteNonQuery -->
        /// <summary>
        ///      Runs the prepared ExecuteNonQuery command
        /// </summary>
        /// <returns></returns>
        public string ExecuteNonQuery()
        {
            _results = "";


            // -------------------------------------------------------------------------90
            //  Open the connection if it isn't already
            // -------------------------------------------------------------------------90
            bool openConnection = _Open();


            // -------------------------------------------------------------------------90
            //  Run stored procedure
            // -------------------------------------------------------------------------90
            int affected = -2;
            try { affected = this.ExecuteNonQuery_simple(); }
            catch (Exception ex)
                { _errors += _.Affix("\r\n\r\nExecuteNonQuery caught error - ", ex.Message); }


            // -------------------------------------------------------------------------90
            //  Look at results
            // -------------------------------------------------------------------------90
            //if (_spReturns > 100000)
            //    _results += _.Affix("the stored procedure '" + this.CommandText + "'"
            //        + " returns: ", BigNumber.Base16(_spReturns));
            //else
                _results += _.Affix("the stored procedure '" + this.CommandText + "'"
                    + " returns: ", _spReturns);


            // -------------------------------------------------------------------------90
            //  Normal post processing, close the connection if this method opened it
            // -------------------------------------------------------------------------90
            _results += _.Affix("\r\n# rows involved: ", affected);
            _results += _.Affix("\r\nthe execute-non-query returns: ", _rowsAffected);
            if (openConnection)
                Connection.Close();


            // -------------------------------------------------------------------------90
            //  Error post processing
            // -------------------------------------------------------------------------90
            if (!string.IsNullOrEmpty(_errors))
            {
                string xml = _Str("@newXML", ""); // for debugging
                Throws.A(new SqlTypeException(_errors), Throws.Actions, _exceptionAction);
            }


            return _results;
        }
        public string ExecuteNonQuery(Endeme exceptionAction)
        {
            _exceptionAction = exceptionAction;
            return ExecuteNonQuery();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteReader -->
        /// <summary>
        ///      Executes a Transact-SQL statement against the connection, returning the results in
        ///      a SqlDataReader
        /// </summary>
        /// <remarks>
        /// Todo:
        ///      rewrite like ExecuteNonQuery
        /// </remarks>
        /// <returns></returns>
        public SqlDataReader ExecuteReader()
        {
            // -------------------------------------------------------------------------90
            //  Pre processing and validation
            // -------------------------------------------------------------------------90
            SqlDataReader rdr = null;
           _results = "";
            if (Connection.State == ConnectionState.Closed)
                Throws.A(new DataException("Change your program"
                    + " - connection needs to be opened externally for a reader"));
            if (_statementCompleted)
            {
                Throws.A(new ArgumentException("Please handle _StatementCompleted in the code surrounding the use of the reader."
                    + "\r\n" + "Set this property false when opening the connection"));
                return null;
            }


            // -------------------------------------------------------------------------90
            //  Execute the reader and catch the results
            // -------------------------------------------------------------------------90
            try { rdr = _command.ExecuteReader(); }
            catch (Exception ex)
            {
                _errors += _.Affix("\r\n_ExecuteReader Error - ", ex.Message);
                if (!string.IsNullOrEmpty(ex.Message))
                    _errors += "\r\nException type:  " + ex.GetType().ToString();
            }
            finally
            {
                if (rdr == null)
                    Throws.A(new DataException("SqlCommand reader is null"), Throws.Actions, _exceptionAction);
                else _results += "reader does " + rdr.HasRows.ToString() + " have rows";
            }


            // -------------------------------------------------------------------------90
            //  Post error processing
            // -------------------------------------------------------------------------90
            if (_errors != "")
                Throws.A(new DataException(_errors), Throws.Actions, _exceptionAction);


            return rdr;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteScalar -->
        /// <summary>
        ///      Runs the prepared ExecuteScalar command
        /// </summary>
        /// <returns></returns>
        public object ExecuteScalar(object defaultValue)
        {
            bool ifOpenedHere = _Open();  // Open the connection if it isn't


            // -------------------------------------------------------------------------90
            //  Run stored procedure
            // -------------------------------------------------------------------------90
            object value = defaultValue;
            _results = "";
            try { value = ExecuteScalar_simple(); }
            catch (Exception ex)
                { _errors += _.Affix("\r\nCaught Error - ", ex.Message); Throws.A(ex);  }
            if (!string.IsNullOrEmpty(_errors))
                Throws.A(new Exception(_errors));


            // -------------------------------------------------------------------------90
            //  Look at results
            // -------------------------------------------------------------------------90
            //if (_spReturns > 100000)
            //    _results += _.Affix("\r\nmy stored procedure returns: ", BigNumber.Base16(_spReturns));
            //else
                _results += _.Affix("\r\nmy stored procedure returns: ", _spReturns);

            _results += _.Affix("\r\nmy execute-non-query returns: ", _rowsAffected);
            _results += _errors;
            _Close(ifOpenedHere);  // Close the connection if this method opened it


            return value;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExecuteScalar_simple -->
        /// <summary>
        ///      Executes a Transact-SQL statement against the connection, returning the result
        /// </summary>
        /// <returns></returns>
        private object ExecuteScalar_simple()
        {
            //_command.CommandText = "SELECT count(*) FROM STARHL7.ADTrecord";
            _errors        = "";
            _statementCompleted = false;
            object value = _command.ExecuteScalar();
            _spReturns    = TreatAs.IntValue(_returnValue.Value, -2);
            return value;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Guid -->
        /// <summary>
        ///      Returns the guid value of the named parameter or an empty guid as default
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns></returns>
        public Guid _Guid(string parameter)
        {
            SqlParameter param = (SqlParameter)Parameters[parameter];
            Guid guid = TreatAs.GuidValue(param.Value, Guid.Empty);
            if (guid == Guid.Empty)
            {
                object obj = _command.Parameters[2].Value;
                guid = TreatAs.GuidValue(obj, Guid.Empty);
            }
            return guid;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _InitMessages -->
        /// <summary>
        ///      Dependency: needs to a have the command's connection object instantiated first
        /// </summary>
        private void _InitMessages()
        {
            if (_command == null)
                Throws.A(new NoNullAllowedException("RichSqlCommand._InitMessages error - command is null"));
            if (_command.Connection == null)
                Throws.A(new NoNullAllowedException("RichSqlCommand._InitMessages error - connection is null"));


            _returnValue = new SqlParameter("@Return_Value", SqlDbType.Int);
            _returnValue.Direction = ParameterDirection.ReturnValue;
            _command.Parameters.Add(_returnValue);
            _command.Connection.FireInfoMessageEventOnUserErrors = true;


            _command.Connection.InfoMessage += new SqlInfoMessageEventHandler(_OnInfoMessage);
            _command.StatementCompleted += new StatementCompletedEventHandler(_OnStatementCompleted);
            _command.Disposed += new EventHandler(_OnDisposed);
            _command.Connection.StateChange += new StateChangeEventHandler(_OnStateChange);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _InitVariables -->
        /// <summary>
        ///      Initializes private member variables
        /// </summary>
        private void _InitVariables()
        {
            _statementCompleted = true; // initialize to true so that readers can tell if it is being handled properly
            _rowsAffected = -1;
            _spReturns = -1;
            _errors = "";


            _command = new SqlCommand();
            _command.Connection = new SqlConnection();
            _inputParam = new List<string>();
        }


        protected void _OnDisposed(object o, EventArgs args)
        {
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _OnInfoMessage -->
        /// <summary>
        /// 
        /// </summary>
        protected void _OnInfoMessage(object o, SqlInfoMessageEventArgs args)
        {
            foreach (SqlError err in args.Errors)
            {
                string subMessage = "\r\nInfoMessage:"
                    + "\r\n"
                    + "a.  The "                  + err.Source
                    + " has received a severity " + err.Class
                    + ", state "                  + err.State + " error\r\n"
                    + "b.  Error number "         + err.Number
                    + " on line "                 + err.LineNumber
                    + " of procedure "            + err.Procedure
                    + " on server "               + err.Server + ":\r\n"
                    + "c.  "                      + err.Message;
                subMessage = Regex.Replace(subMessage, "of procedure  on server", "of (command text?) on server");
                _errors += "\r\n" + subMessage;
            }
            _errors += "\r\n\r\nSummary:  "        + args.Message
                    + "\r\nFrom Source:  "        + args.Source;
        }


        protected void _OnStateChange(object o, StateChangeEventArgs args)
        {
            ConnectionState state = args.CurrentState;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _OnStatementCompleted -->
        /// <summary>
        ///      If the statement does not complete, something probably went wrong
        /// </summary>
        protected void _OnStatementCompleted(object o, StatementCompletedEventArgs args)
        {
            _statementCompleted = true;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Open -->
        /// <summary>
        ///      Opens the connection if it is not already open
        /// </summary>
        /// <returns>Side effect: returns whether it opened the connection or not</returns>
        public bool _Open()
        {
            if (Connection.State == ConnectionState.Closed)
            {
                //MessageBox.Show("Before open connection");
                try
                {
                    Connection.Open();
                }
                catch (Exception ex)
                {
                    Type type = ex.GetType();
                    Throws.A(ex);
                }
                //MessageBox.Show("After open connection");
                return true;
            }
            else return false;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _RemoveInputParameters -->
        /// <summary>
        ///      Removes all the parameters except '@Return_Value'
        /// </summary>
        private void _RemoveParametersExcept(string paramName)
        {
            for (int i = Parameters.Count - 1; i >= 0; --i)
            {
                SqlParameter param = (SqlParameter)Parameters[i];
                if (param.ParameterName != paramName) Parameters.RemoveAt(1);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _SetParam -->
        /// <summary>
        ///      Sets the value of a parameter whose type is already set
        /// </summary>
        /// <param name="paramName"></param>
        /// <param name="paramType"></param>
        /// <param name="paramValue"></param>
        public void _SetParam(string paramName, object paramValue)
        {
            _command.Parameters[paramName].Value = paramValue;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- _Str -->
        /// <summary>
        ///      Returns the value of a stored preocedure parameter as a string
        /// </summary>
        /// <param name="parameterName"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public string _Str(string parameterName, string defaultValue)
        {
            string str;
            if (Parameters.Contains(parameterName))
                str = TreatAs.StrValue(((SqlParameter)Parameters[parameterName]).Value, defaultValue);
            else
                str = defaultValue;
            return str;
        }


        #endregion new and rewritten methods

    }
}
