// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;                       // for Regex

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- Characteristic -->
    /// <summary>
    ///      Endemes and EndemeSets are made up of Characteristics
    /// </summary>
    public class EndemeCharacteristic
    {
        // ----------------------------------------------------------------------------------------
        //  Members
        // ----------------------------------------------------------------------------------------
        public char   Letter   { get { return _letter;   } } private char   _letter;
        public string Label    { get { return _label;    } } private string _label;
        public string Descr    { get { return _descr;    } } private string _descr;
        public string Units    { get { return _units;    } } private string _units;
        public double Value    { get { return _value;    } } private double _value;
        public char   Opposite { get { return _opposite; } } private char   _opposite;


        // ----------------------------------------------------------------------------------------
        /// <!-- PositionMeaning -->
        /// <summary>
        ///      Stores what each value of a characteristic means
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks>should probably just make this private and build accessors</remarks>
        public List<EndemeMeaning> PositionMeaning
        {
            get { return _positionMeaning; }
        }
        private List<EndemeMeaning> _positionMeaning;


        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        public EndemeCharacteristic(char cha, string label, string descr)
        {
            Set(cha, label, descr, "", 0.0, ' ');
        }
        public EndemeCharacteristic(char cha, string label, string descr, string units, double value, char opposite)
        {
            Set(cha, label, descr, units, value, opposite);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Set -->
        /// <summary>
        ///      Sets the value of a characteristic
        /// </summary>
        /// <param name="cha"></param>
        /// <param name="label"></param>
        /// <param name="descr"></param>
        /// <param name="units"></param>
        /// <param name="value"></param>
        private void Set(char cha, string label, string descr, string units, double value, char opposite)
        {
            _letter   = cha;
            _label    = label;
            _descr    = descr;
            _units    = units;
            _value    = value;
            _opposite = opposite;
        }
        public void Set(char cha, string label, string descr) { Set(cha, label, descr, "", 0.0, ' '); }


        // ----------------------------------------------------------------------------------------
        /// <!-- Display -->
        /// <summary>
        /// 
        /// </summary>
        public string Display { get
        {
            if (string.IsNullOrWhiteSpace(_label) || _label.ToCharArray()[0] == _letter)
                return Regex.Replace(_label, "^(.)", "$1)");
            else return "" + _letter + ")" + _label;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return Letter.ToString() + ". " + Label + " - " + Descr;
        }
    }
}
