﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;   // for Regex

namespace InformationLib.Endemes
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeHelpers -->
    public static class EndemeHelpers
    {
        // ----------------------------------------------------------------------------------------
        /// <!-- MoveLetter -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="str">string to modify</param>
        /// <param name="letter">string (letter) to move</param>
        /// <param name="toPosition">1 to n or -1 to -n</param>
        /// <returns></returns>
        public static string MoveLetter(string str, string letter, int toPosition)
        {
            //  remove the letter
            string s2 = Regex.Replace(str, letter, "");

            if (toPosition < 0)
            {
                toPosition = -toPosition;
                toPosition--; // convert 1 to N to 0 to n-1
                s2 = Regex.Replace(s2, "(.{" + toPosition + "})$", letter + "$1");
            }
            else
            {
                toPosition--; // convert 1 to N to 0 to n-1
                s2 = Regex.Replace(s2, "^(.{" + toPosition + "})", "$1" + letter);
            }

            return s2;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Best -->
        /// <summary>
        ///      Return an endeme no matter what
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue">the set will be used if the object is not an Endeme</param>
        /// <returns></returns>
        public static Endeme Best(Object obj, Endeme defaultValue)
        {
            if (obj == null)                      return defaultValue.Copy();
            if (obj.GetType() == typeof(Endeme))  return ((Endeme)obj).Copy();
            String str = Regex.Replace(obj.ToString(), "[^A-Za-z]", "");
            if (defaultValue == null)
            {
                if (string.IsNullOrEmpty(str))    return new Endeme(new EndemeSet());
                else                              return new Endeme(new EndemeSet(), str);
            }
            else
            {
                if (string.IsNullOrEmpty(str))    return defaultValue.Copy();
                else                              return new Endeme(defaultValue.Set, str);
            }
        }
    }
}
