﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeItem -->
    /// <summary>
    ///      The EndemeItem class contains the endeme and its related data (and information)
    /// </summary>
    /// <remarks>
    ///                     EndemeItem (endeme leaf in the component pattern)
    ///      EndemeSet          |
    ///          ^              +-------+----+----+
    ///          |              |       |    |    |
    ///          +----------- Endeme   Value | Number
    ///                         |          Text
    ///                 char -> |
    ///                         |
    ///                    Raw, Cooked
    /// </remarks>
    public class EndemeItem : EndemeAbstractComponent
    {

        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme -->
        /// <summary>
        ///      The main event
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks></remarks>
        public override Endeme Endeme { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Members for indexing lists etc
        // ----------------------------------------------------------------------------------------
        /// <summary>The label of the item</summary>
        public override string Label { get; set; }              /// <summary>The Guid key of the node or leaf</summary>
        public override Guid   Key   { get; set; }              /// <summary>The raw strength of the match</summary>
        public          int    Match { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Members for storing data in an information context
        // ----------------------------------------------------------------------------------------
        // mismatch rate
        // mismatch threshold
        // number of data items expected
        public override object Value { get; set; }
        public override double Number { get; set; }
        public override string Text { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        public EndemeItem(         Endeme endeme)                { Key = Guid.NewGuid(); Endeme = endeme;                  Label = "";    }
        public EndemeItem(         string endeme, EndemeSet set) { Key = Guid.NewGuid(); Endeme = new Endeme(set, endeme); Label = "";    }
        public EndemeItem(Guid id, Endeme endeme)                { Key = id;             Endeme = endeme;                  Label = "";    }
        public EndemeItem(Guid id, Endeme endeme, string label)  { Key = id;             Endeme = endeme;                  Label = label; }
        public EndemeItem(EndemeItem item)
        {
            Key    = item.Key;
            Endeme = item.Endeme; //.Copy() copying a reference, rather than a value
            Label  = item.Label;
            Number = item.Number;
            Text   = item.Text;
            Value  = item.Value; // copying a reference, rather than a value
            Match  = item.Match;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Copy -->
        /// <summary>
        ///      Returns a copy of the EndemeItem where the Value and the Set are references
        /// </summary>
        /// <returns></returns>
        public override EndemeAbstractComponent Copy()
        {
            EndemeItem item = new EndemeItem(this.Key, this.Endeme.Copy(), this.Label);
            item.Match  = this.Match;
            item.Value  = this.Value;                // copying a reference, rather than a value
            item.Number = this.Number;
            item.Text   = this.Text;
            return item;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Count -->
        /// <summary>
        ///      Returns the count of leaves, in this case, 1
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks></remarks>
        public override int Count
        {
            get { return 1; }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IsEmpty -->
        /// <summary>
        ///      Determines whether the endeme field is empty or blank
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks></remarks>
        public override bool IsEmpty
        {
            get
            {
                if (Endeme == null || Endeme.Count == 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        ///      Default string display
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            string str = "";
            if (Endeme != null) str = Endeme.ToString() + " <" + Match + "> " + Label + " " + Number;
            else str = "NULL" + " <" + Match + "> " + Label + " " + Number;
            if (Value == null) return str;
            else return str + ", " + Value.ToString();
        }


        //public bool IsEmpty { get { if (Endeme == null || Endeme.Count == 0) return true; else return false; } }
    }
}
