﻿// (C) Jon Grover, 2012
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeItem -->
    /// <summary>
    ///      The EndemeItem class contains the endeme and its related data (and information)
    /// </summary>
    /// <remarks>
    ///                     EndemeItem (endeme leaf in the component pattern)
    ///      EndemeSet          |
    ///          ^              +-------+----+----+
    ///          |              |       |    |    |
    ///          +----------- Endeme   Value | Number
    ///                         |          Text
    ///                 char -> |
    ///                         |
    ///                    Raw, Cooked
    /// </remarks>
    public class EndemeItem
    {
        public Endeme Endeme { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Members for indexing lists etc
        // ----------------------------------------------------------------------------------------
        public string Label  { get; set; }
        public Guid   Key    { get; set; }
        public int    Match  { get; set; }  // the raw strength of the match


        // ----------------------------------------------------------------------------------------
        //  Members for storing data as information
        // ----------------------------------------------------------------------------------------
        // mismatch rate
        // mismatch threshold
        // number of data items expected
        public object Value  { get; set; }
        public double Number { get { return _number; } set { _number = value; } } private double _number;
        public string Text   { get { return _text;   } set { _text   = value; } } private string _text;


        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        public EndemeItem(         Endeme endeme)                { Key = Guid.NewGuid(); Endeme = endeme;                  Label = "";    }
        public EndemeItem(         string endeme, EndemeSet set) { Key = Guid.NewGuid(); Endeme = new Endeme(set, endeme); Label = "";    }
        public EndemeItem(Guid id, Endeme endeme)                { Key = id;             Endeme = endeme;                  Label = "";    }
        public EndemeItem(Guid id, Endeme endeme, string label)  { Key = id;             Endeme = endeme;                  Label = label; }
        public EndemeItem(EndemeItem item)
        {
            Key    = item.Key;
            Endeme = item.Endeme; //.Copy() copying a reference, rather than a value
            Label  = item.Label;
            Number = item.Number;
            Text   = item.Text;
            Value  = item.Value; // copying a reference, rather than a value
            Match  = item.Match;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Copy -->
        /// <summary>
        ///      Returns a copy of the EndemeItem where the Value and the Set are references
        /// </summary>
        /// <returns></returns>
        public EndemeItem Copy()
        {
            EndemeItem item = new EndemeItem(this.Key, this.Endeme.Copy(), this.Label);
            item.Match  = this.Match;
            item.Value  = this.Value; // copying a reference, rather than a value
            item.Number = this.Number;
            item.Text   = this.Text;
            return item;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        ///      Default string display
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            string str = "";
            if (Endeme != null) str = Endeme.ToString() + " <" + Match + "> " + Label + " " + Number;
            else str = "NULL" + " <" + Match + "> " + Label + " " + Number;
            if (Value == null) return str;
            else return str + ", " + Value.ToString();
        }


        public bool IsEmpty { get { if (Endeme == null || Endeme.Count == 0) return true; else return false; } }
    }
}
