﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InformationLib.Endemes
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeNodeConnector_old -->
    /// <summary>
    ///      The EndemeNodeConnector_old class contains a number of endeme lists, one for each EndemeSet
    ///      name, each list contains a loosely coupled group of fields/columns/members, this
    ///      probably has some advantages over the tight coupling of fields in an object or table
    /// </summary>
    /// <remarks>
    ///      Warning - this version of EndemeNodeConnector is deprecated
    ///      
    /// 
    ///                                 EndemeNode <---------------------+   (endeme info item)
    ///                                      |                           |
    ///                    EndemeSet ------> |                           |
    ///                     |     ^          |                           |
    ///                     |     |         /|\                          |
    ///                     |     +---- EndemeArray                      |
    ///                     |     |          |                           |
    ///                     |     +--------> | <-- label, idx, Guid      |
    ///                     |     |         /|\  \   \    \    \    \    |
    ///                     |     |      EndemeLeaf - - - - - - - - EndemeNode (Remove, For-(Add,get child), composite?)
    ///                     |     |          |     \               /                       
    ///                     |     |          |     IEndemeComponent  (Endeme, Guid)        
    ///                     |     |          |                            
    ///                     |     |          +---------+-------+------+------+
    ///                     |     |          |         |       |      |      |
    ///                     |     +------- Endeme   Number   Text   Value   Guid
    ///                     |                ||                       |
    ///                     | <--- char ---> ||                   (anything)
    ///                    /|\              /||\
    ///       EndemeCharacteristic <--> EndemeQuantification
    ///       
    ///                        node                                    node                    ? leaf
    ///       i=2 set=[person] field=[physiciane]  ====>  j=2 set=[set] field=[endeme]  ====>  number=3
    /// </remarks>
    public class EndemeNodeConnector_old
    {
        // ----------------------------------------------------------------------------------------
        //  Members
        // ----------------------------------------------------------------------------------------
        Dictionary<string, EndemeList> _dictionary;


        // ----------------------------------------------------------------------------------------
        //  Accessors
        // ----------------------------------------------------------------------------------------
        public EndemeList this[string set]  { get { return _dictionary[set]; } }


        // ----------------------------------------------------------------------------------------
        //  Constructor
        // ----------------------------------------------------------------------------------------
        public EndemeNodeConnector_old()
        {
            _dictionary = new Dictionary<string, EndemeList>();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EvinceList -->
        /// <summary>
        ///      Returns the requested list, creating it if it does not yet exist
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public EndemeList EvinceList(EndemeSet set)
        {
            if (!_dictionary.ContainsKey(set.Name))
                _dictionary.Add(set.Name, new EndemeList(1, set));
            EndemeList list = _dictionary[set.Name];
            return list;
        }

    }
}
