﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeNode -->
    /// <summary>
    ///      The EndemeNode class contains a collection by 1) int/char, 2) endeme/context,
    ///      and 3) endeme set name and is used for storing soft coded variables, members, and fields
    /// </summary>
    /// <remarks>
    ///      Warning - this version of EndemeNode is deprecated.
    ///      
    /// 
    ///      This is my first try at making complex large structures with endemes, 
    ///      I designed it after the segment in HL7, and it should be able to contain
    ///      one HL7 segment, or possibly the entire HL7.
    ///      
    ///      Actually since I wrote that, I have decided to use the composite pattern:
    /// 
    /// 
    ///                                   EndemeNode <---------------------+   (endeme info item)
    ///                                        |                           |                          
    ///                          int/char ---> |                           |
    ///                                       /|\                          |
    ///                               EndemeNodeConnector                  |
    ///                                        |                           |
    ///                    EndemeSet --------> |                           |
    ///                     |     ^            |                           |
    ///                     |     |           /|\                          |
    ///                     |     +------ EndemeList                       |
    ///                     |     |            |                           |
    ///                     |     +----------> | <-- label, idx, Guid      |
    ///                     |     |            |                           |
    ///                     |     |            |      IEndemeComponent     |
    ///                     |     |           /|\    /               \     |
    ///                     |     | (Leaf) EndemeItem - - - - - - - - EndemeNode (composite?)
    ///                     |     |            |                            
    ///                     |     |            +---------+-------+------+------+
    ///                     |     |            |         |       |      |      |
    ///                     |     +--------- Endeme   Number   Text   Value   Guid
    ///                     |                  ||                       |
    ///                     | <---- char ----> ||                   (anything)
    ///                    /|\                /||\
    ///       EndemeCharacteristic <---> EndemeQuantification
    /// 
    /// 
    ///             node           leaf                        node        leaf
    ///       i=2 set=[person] field=[physician]  ====>  j=2 set=[set] field=[endeme]  ====>  number=3
    /// </remarks>
    public class EndemeNode_old
    {
        // ----------------------------------------------------------------------------------------
        //  Members
        // ----------------------------------------------------------------------------------------
        public Dictionary<int, EndemeNodeConnector_old> _valueGroup { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Accessors - custom indexers
        // ----------------------------------------------------------------------------------------
        public EndemeList this[char key, EndemeSet set] { get { return _valueGroup[(int)key][set.Name]; } }
        public EndemeList this[int  key, EndemeSet set] { get { return _valueGroup[     key][set.Name]; } }


        // ----------------------------------------------------------------------------------------
        //  Constructor
        // ----------------------------------------------------------------------------------------
        public EndemeNode_old(int capacity, EndemeSet set) 
        {
            _valueGroup = new Dictionary<int, EndemeNodeConnector_old>(capacity);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Average -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="endeme"></param>
        /// <param name="set"></param>
        /// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
        /// <returns></returns>
        public double Average(string endeme, EndemeSet set, double closeness)
        {
            double total = 0.0;
            foreach (char key in _valueGroup.Keys)
            {
                Endeme e = From(key, endeme, set, closeness).Endeme;
                total += From(key, endeme, set, closeness).Number;
            }
            return (total/_valueGroup.Count);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EvinceGroup -->
        /// <summary>
        ///      Returns the requested list, creating it if it does not yet exist
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public EndemeNodeConnector_old EvinceGroup(char key)
        {
            if (!_valueGroup.ContainsKey(key))
                _valueGroup.Add(key, new EndemeNodeConnector_old());
            EndemeNodeConnector_old dictionary = _valueGroup[key];
            return dictionary;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- For -->
        /// <summary>
        ///      Gets the exact data context, this may not be what you want if you DON'T want to create a new one
        /// </summary>
        /// <param name="key"></param>
        /// <param name="endeme"></param>
        /// <param name="set"></param>
        /// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
        /// <returns></returns>
        public EndemeItem For(char key, string endeme, EndemeSet set, double closeness)
        {
            EndemeNodeConnector_old group = EvinceGroup(key);
            EndemeList list = group.EvinceList(set);
            EndemeItem item = (EndemeItem)list.EvinceItem(endeme);
            return item;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- From -->
        /// <summary>
        ///      Returns the best matching data context, this may not be what you want of you DO want to create a new one
        /// </summary>
        /// <remarks>
        ///      To add - rarely return null, try a best context approximation, include maybe a match threshold
        /// </remarks>
        /// <param name="key"></param>
        /// <param name="endeme"></param>
        /// <param name="set"></param>
        /// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
        /// <returns></returns>
        public EndemeItem From(char key, string endeme, EndemeSet set, double closeness)
        {
            EndemeNodeConnector_old group = EvinceGroup(key);
            EndemeList list = group.EvinceList(set);
            EndemeItem item = (EndemeItem)list[new Endeme(set, endeme)];
            if (item != null) return item;
            else return null;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StdDeviation -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="endeme"></param>
        /// <param name="set"></param>
        /// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
        /// <returns></returns>
        public double StdDeviation(string endeme, EndemeSet set, double closeness)
        {
            List<double> list = new List<double>();
            foreach (char key in _valueGroup.Keys)
                list.Add(From(key, endeme, set, closeness).Number);
            return StdDeviation(list);
        }
        public static double StdDeviation(List<double> list)
        {
            if (list == null || list.Count == 0) return 0.0;
            double average = list.Average();
            double sumOfDifferencesSquared = list.Select(val => ((double)val - average) * ((double)val - average)).Sum();
            double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
            return sd;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToDictionary -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="endeme"></param>
        /// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
        /// <returns></returns>
        public Dictionary<char, int> ToDictionary(string endeme, EndemeSet set, double closeness)
        {
            Dictionary<char, int> dictionary = new Dictionary<char, int>(_valueGroup.Count);
            foreach (char key in _valueGroup.Keys)
                dictionary.Add(key, (int)From(key, endeme, set, closeness).Number);
            return dictionary;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "" + _valueGroup.Count;
        }

    }


}
