// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using InformationLib.SoftData;
using System;                                     // for Random
using System.Collections.Generic;                 // for List
using System.Text;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeTests -->
    /// <summary>
    ///      The EndemeTests class tests the classes in the Endemes Library
    /// </summary>
    /// <remarks>
    /// </remarks>
    //[TestFixture]
    public class EndemeTests
    {
        private string _result;
        private static string test1 = "\r\n((((((((((   ";
        private static string test2 = " test   )))))))))) ";


        // ----------------------------------------------------------------------------------------
        /// <!-- AllTests -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string AllTests()
        {
            _result = "";
            string result = "";


            // --------------------------------------------------------------------------
            //  Constructor tests
            // --------------------------------------------------------------------------
            Endeme_Constructor_test          ();  result += Affix(test1 + "Endeme Constructor     " + test2, _result);
            EndemeCharacteristic_Constructor_test();  result += Affix(test1 + "EndemeChar Constructor " + test2, _result);
            EndemeLeaf_Constructor_test      ();  result += Affix(test1 + "EndemeLeaf constructor " + test2, _result); // do value first and value last have the correct items
            EndemeList_Constructor_test      ();  result += Affix(test1 + "EndemeList constructor " + test2, _result); // do value first and value last have the correct items
            EndemeNormalizer_Constructor_test();  result += Affix(test1 + "EndemeNorm constructor " + test2, _result);
            EndemeSet_Constructor_test       ();  result += Affix(test1 + "EndemeSet constructor  " + test2, _result); // is the len of the random endeme = 22


            // --------------------------------------------------------------------------
            //  Endeme Method tests
            // --------------------------------------------------------------------------
            EndemeSet_ToString_test          ();  result += Affix(test1 + "EndemeSet ToString     " + test2, _result);
            Endeme_Show_test                 ();  result += Affix(test1 + "Endeme Show            " + test2, _result);
            Endeme_BuildString_test          ();  result += Affix(test1 + "Endeme BuildString     " + test2, _result);
            Endeme_Bag_test                  ();  result += Affix(test1 + "Endeme Bag             " + test2, _result);

            EndemeList_StdDeviation_test     ();  result += Affix(test1 + "EndemeList StdDeviation" + test2, _result);
            EndemeHelpers_MoveLetter_test    ();  result += Affix(test1 + "EndemeHelper MoveLetter" + test2, _result);


            Range_Wrap_test                  ();  result += Affix(test1 + "Range Wrap             " + test2, _result);
            //Endeme_MatchAverage_test  ();  result += Affix(test1 + "Endeme average       " + test2, _result);
            //Endeme_Match_test         ();  result += Affix(test1 + "Endeme match         " + test2, _result);
            //Endeme_MatchOrder_test    ();  result += Affix(test1 + "Match Order          " + test2, _result);


            result += "\r\nendeme tests succeeded";
            return result;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme_Bag_test -->
        /// <summary>
        ///      Tests the Bag method in the Endeme class
        /// </summary>
        //[Test]
        private void Endeme_Bag_test()
        {
            _result = "";
            Endeme endeme = new Endeme(WetlandAnimals, "RPIQLTSJAOUDBCFEHNKVMG");
            EndemeDrawBag bag;
            bag = new EndemeDrawBag(endeme, 64);  Assert.That(bag.Count == 64);
            bag = new EndemeDrawBag(endeme, 32);  Assert.That(bag.Count == 32);
            bag = new EndemeDrawBag(endeme, 16);  Assert.That(bag.Count == 16);
            bag = new EndemeDrawBag(endeme,  8);  Assert.That(bag.Count ==  8);
            bag = new EndemeDrawBag(endeme,  4);  Assert.That(bag.Count ==  4);
            bag = new EndemeDrawBag(endeme,  2);  Assert.That(bag.Count ==  2);
            bag = new EndemeDrawBag(endeme,  1);  Assert.That(bag.Count ==  1);
            _result = "OK";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Range_Wrap_test -->
        /// <summary>
        ///      Tests the Wrap method in the Range class
        /// </summary>
        //[Test]
        public void Range_Wrap_test()
        {
            Range_Wrap_testcase(0, 5, true, false, -3, 2);
            Range_Wrap_testcase(0, 5, true, false, -2, 3);
            Range_Wrap_testcase(0, 5, true, false, -1, 4);
            Range_Wrap_testcase(0, 5, true, false,  0, 0);
            Range_Wrap_testcase(0, 5, true, false,  1, 1);
            Range_Wrap_testcase(0, 5, true, false,  2, 2);
            Range_Wrap_testcase(0, 5, true, false,  3, 3);
            Range_Wrap_testcase(0, 5, true, false,  4, 4);
            Range_Wrap_testcase(0, 5, true, false,  5, 0);
            Range_Wrap_testcase(0, 5, true, false,  6, 1);
            Range_Wrap_testcase(0, 5, true, false,  7, 2);
            Range_Wrap_testcase(0, 5, true, false,  8, 3);

            Range_Wrap_testcase(0, 5, false, false, -3, 1);
            Range_Wrap_testcase(0, 5, false, false, -2, 2);
            Range_Wrap_testcase(0, 5, false, false, -1, 3);
            Range_Wrap_testcase(0, 5, false, false,  0, 4);
            Range_Wrap_testcase(0, 5, false, false,  1, 1);
            Range_Wrap_testcase(0, 5, false, false,  2, 2);
            Range_Wrap_testcase(0, 5, false, false,  3, 3);
            Range_Wrap_testcase(0, 5, false, false,  4, 4);
            Range_Wrap_testcase(0, 5, false, false,  5, 1);
            Range_Wrap_testcase(0, 5, false, false,  6, 2);
            Range_Wrap_testcase(0, 5, false, false,  7, 3);
            Range_Wrap_testcase(0, 5, false, false,  8, 4);

            Range_Wrap_testcase(0, 5, true, true, -3, 3);
            Range_Wrap_testcase(0, 5, true, true, -2, 4);
            Range_Wrap_testcase(0, 5, true, true, -1, 5);
            Range_Wrap_testcase(0, 5, true, true,  0, 0);
            Range_Wrap_testcase(0, 5, true, true,  1, 1);
            Range_Wrap_testcase(0, 5, true, true,  2, 2);
            Range_Wrap_testcase(0, 5, true, true,  3, 3);
            Range_Wrap_testcase(0, 5, true, true,  4, 4);
            Range_Wrap_testcase(0, 5, true, true,  5, 5);
            Range_Wrap_testcase(0, 5, true, true,  6, 0);
            Range_Wrap_testcase(0, 5, true, true,  7, 1);
            Range_Wrap_testcase(0, 5, true, true,  8, 2);

            Range_Wrap_testcase(0, 5, false, true, -3, 2);
            Range_Wrap_testcase(0, 5, false, true, -2, 3);
            Range_Wrap_testcase(0, 5, false, true, -1, 4);
            Range_Wrap_testcase(0, 5, false, true,  0, 5);
            Range_Wrap_testcase(0, 5, false, true,  1, 1);
            Range_Wrap_testcase(0, 5, false, true,  2, 2);
            Range_Wrap_testcase(0, 5, false, true,  3, 3);
            Range_Wrap_testcase(0, 5, false, true,  4, 4);
            Range_Wrap_testcase(0, 5, false, true,  5, 5);
            Range_Wrap_testcase(0, 5, false, true,  6, 1);
            Range_Wrap_testcase(0, 5, false, true,  7, 2);
            Range_Wrap_testcase(0, 5, false, true,  8, 3);

            Range_Wrap_testcase(2, 7, true, false, -3, 2);
            Range_Wrap_testcase(2, 7, true, false, -2, 3);
            Range_Wrap_testcase(2, 7, true, false, -1, 4);
            Range_Wrap_testcase(2, 7, true, false,  0, 5);
            Range_Wrap_testcase(2, 7, true, false,  1, 6);
            Range_Wrap_testcase(2, 7, true, false,  2, 2);
            Range_Wrap_testcase(2, 7, true, false,  3, 3);
            Range_Wrap_testcase(2, 7, true, false,  4, 4);
            Range_Wrap_testcase(2, 7, true, false,  5, 5);
            Range_Wrap_testcase(2, 7, true, false,  6, 6);
            Range_Wrap_testcase(2, 7, true, false,  7, 2);
            Range_Wrap_testcase(2, 7, true, false,  8, 3);

            Range_Wrap_testcase(2, 7, false, false, -3, 5);
            Range_Wrap_testcase(2, 7, false, false, -2, 6);
            Range_Wrap_testcase(2, 7, false, false, -1, 3);
            Range_Wrap_testcase(2, 7, false, false,  0, 4);
            Range_Wrap_testcase(2, 7, false, false,  1, 5);
            Range_Wrap_testcase(2, 7, false, false,  2, 6);
            Range_Wrap_testcase(2, 7, false, false,  3, 3);
            Range_Wrap_testcase(2, 7, false, false,  4, 4);
            Range_Wrap_testcase(2, 7, false, false,  5, 5);
            Range_Wrap_testcase(2, 7, false, false,  6, 6);
            Range_Wrap_testcase(2, 7, false, false,  7, 3);
            Range_Wrap_testcase(2, 7, false, false,  8, 4);

            Range_Wrap_testcase(2, 7, true, true, -3, 3);
            Range_Wrap_testcase(2, 7, true, true, -2, 4);
            Range_Wrap_testcase(2, 7, true, true, -1, 5);
            Range_Wrap_testcase(2, 7, true, true,  0, 6);
            Range_Wrap_testcase(2, 7, true, true,  1, 7);
            Range_Wrap_testcase(2, 7, true, true,  2, 2);
            Range_Wrap_testcase(2, 7, true, true,  3, 3);
            Range_Wrap_testcase(2, 7, true, true,  4, 4);
            Range_Wrap_testcase(2, 7, true, true,  5, 5);
            Range_Wrap_testcase(2, 7, true, true,  6, 6);
            Range_Wrap_testcase(2, 7, true, true,  7, 7);
            Range_Wrap_testcase(2, 7, true, true,  8, 2);

            Range_Wrap_testcase(2, 7, false, true, -3, 7);
            Range_Wrap_testcase(2, 7, false, true, -2, 3);
            Range_Wrap_testcase(2, 7, false, true, -1, 4);
            Range_Wrap_testcase(2, 7, false, true,  0, 5);
            Range_Wrap_testcase(2, 7, false, true,  1, 6);
            Range_Wrap_testcase(2, 7, false, true,  2, 7);
            Range_Wrap_testcase(2, 7, false, true,  3, 3);
            Range_Wrap_testcase(2, 7, false, true,  4, 4);
            Range_Wrap_testcase(2, 7, false, true,  5, 5);
            Range_Wrap_testcase(2, 7, false, true,  6, 6);
            Range_Wrap_testcase(2, 7, false, true,  7, 7);
            Range_Wrap_testcase(2, 7, false, true,  8, 3);
        }
        private void Range_Wrap_testcase(double low, double high, bool lowInc, bool highInc, int input, int target)
        {
            Range range = new Range(low, high, lowInc, highInc);
            int value = range.Wrap( input);
            Assert.That(value == target);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme_BuildString_test -->
        /// <summary>
        ///      Tests BuildString, Also tests BuildIndex since these two methods are closely associated
        /// </summary>
        //[Test]
        public void Endeme_BuildString_test()
        {
            //  Test BuildString and BuildIndex without an endeme set involved
            Endeme_BuildString_testcase();

            //  Test BuildString and BuildIndex with an endeme set involved
            Endeme_BuildString_testcase(WetlandAnimals);
        }
        private void Endeme_BuildString_testcase(EndemeSet WetlandAnimals)
        {
            // create an empty endeme of WetlandAnimals
            Endeme e = new Endeme(WetlandAnimals);


            // use BuildString and the other std initializers
            string endeme = "ABCDCDEFGHIJKLM";
            e.BuildString(endeme);
            e.BuildIndex();
            //e.InitRaw();


            // check the index using a number
            int len = e.Count / 2;
            char c = e[len];
            int idx = e.Size - e[c];
            Assert.That(idx == len);


            // check the index using a character
            char j = 'J';
            int idxj = e.Size - e[j];
            char j2 = e[idxj];
            Assert.That(j2 == j);


            // test to see that there are not duplicate characters
            Dictionary<char, int> test = new Dictionary<char, int>();
            for (int i = 0; i < e.Count; ++i)
            {
                char ctest = e[i];
                if (test.ContainsKey(ctest))
                    test[ctest] = 2;
                else test.Add(ctest, 1);
            }
            Assert.That(!test.ContainsValue(2));
        }
        public void Endeme_BuildString_testcase()
        {
            // create an empty endeme
            Endeme e = new Endeme();


            // use BuildString and the other std initializers
            string endeme = "ABCDCDEFGHIJKLM";
            e.BuildString(endeme);
            e.BuildIndex();
            //e.InitRaw();


            // check the index using a number
            int len = e.Count / 2;
            char c = e[len];
            int idx = e.Size - e[c];
            Assert.That(idx == len);


            // check the index using a character
            char j = 'J';
            int idxj = e.Size - e[j];
            char j2 = e[idxj];
            Assert.That(j2 == j);


            // test to see that there are not duplicate characters
            Dictionary<char, int> test = new Dictionary<char, int>();
            for (int i = 0; i < e.Count; ++i)
            {
                char ctest = e[i];
                if (test.ContainsKey(ctest))
                    test[ctest] = 2;
                else test.Add(ctest, 1);
            }
            Assert.That(!test.ContainsValue(2));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeHelpers_MoveLetter_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void EndemeHelpers_MoveLetter_test()
        {
            _result = "";
            string str;
            str = EndemeHelpers.MoveLetter("hi", "i",  1); Assert.That(str == "ih");
            str = EndemeHelpers.MoveLetter("hi", "i",  2); Assert.That(str == "hi");
            str = EndemeHelpers.MoveLetter("hi", "i", -1); Assert.That(str == "hi");
            str = EndemeHelpers.MoveLetter("hi", "i", -2); Assert.That(str == "ih");
            _result = "OK";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeList_StdDeviation_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void EndemeList_StdDeviation_test()
        {
            //  Test the general std deviation algorithm
            List<double> numList = new List<double>(8);
            numList.Add(2.0);
            numList.Add(4.0);
            numList.Add(4.0);
            numList.Add(4.0);
            numList.Add(5.0);
            numList.Add(5.0);
            numList.Add(7.0);
            numList.Add(9.0);
            double num = EndemeList.StdDeviation(numList);
            Assert.That(num == 2.0);


            //  Test an empty list
            List<Endeme> endeme = new List<Endeme>(4);
            EndemeList list = new EndemeList(12, WetlandAnimals);
            Endeme stdNone = list.StdDeviation();
            Assert.That(stdNone.Count == 0);


            //  Test a single item list
            endeme.Add(new Endeme(WetlandAnimals, "TUKHQDNVILGRSOCBFAJPME"));  endeme[0].GenerateRaw_geometric(0.8);  list.Add(endeme[0], null);
            Endeme stdOne = list.StdDeviation();
            Assert.That(stdOne.Quant.Raw['C'] == 0.0);


            endeme.Add(new Endeme(WetlandAnimals, "TMVLJNFGSUKBPOEDHCIQRA"));  endeme[1].GenerateRaw_geometric(0.8);  list.Add(endeme[1], null);
            endeme.Add(new Endeme(WetlandAnimals, "NTSVPCGUOKARBDLIHFJQME"));  endeme[2].GenerateRaw_geometric(0.8);  list.Add(endeme[2], null);
            endeme.Add(new Endeme(WetlandAnimals, "NGCPVBUIEDHFKALTOSQJRM"));  endeme[3].GenerateRaw_geometric(0.8);  list.Add(endeme[3], null);
            Endeme std = list.StdDeviation();
            Assert.That(17.008 < std.Quant.Raw['J'] && std.Quant.Raw['J'] < 17.009);
            Endeme avg = list.Average();
            Assert.That(11.501 < avg.Quant.Raw['J'] && avg.Quant.Raw['J'] < 11.502);


            //Endeme other = list.StdDeviation(); //list.StdDeviations();
            //Endeme avgs = list.Averages();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StoreMatch_pair -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="distrib"></param>
        /// <param name="keys"></param>
        /// <param name="rawMatch"></param>
        private static void StoreMatch_pair(Dictionary<int, int> distrib, List<int> keys, int rawMatch)
        {
            if (distrib.ContainsKey(rawMatch)) distrib[rawMatch]++;
            else
            {
                distrib.Add(rawMatch, 1);
                keys.Add(rawMatch);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme_MatchAverage_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void Endeme_MatchAverage_test()
        {
            EndemeTestTally tally = new EndemeTestTally(100);
            EndemeSet animals = WetlandAnimals;
            Random r = new Random();


            for (int i = 0; i < 100; ++i)
            {
                Endeme e1 = animals.RandomEndeme(r);
                Endeme e2 = animals.RandomEndeme(r);
                double match = animals.Match(e1, e2);
                tally.Incr((int)(match * 99.0 + 1.0), 2);
            }
            double avg = tally.Average();
            Assert.That(45.0 < avg && avg < 55.0);


            _result = avg.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void Endeme_Constructor_test()
        {
            _result = "";
            EndemeConstructor_testcase1();

            EndemeConstructor_testcase2("ABCDEFGHIJKLMNOPQRSTUV", 22, "ABCDEFGHIJKLMNOPQRSTUV");
            EndemeConstructor_testcase2("abcdefghijklmnopqrstuv", 22, "ABCDEFGHIJKLMNOPQRSTUV");
            EndemeConstructor_testcase2("AABCDEFGHIJKLMNOPQRSTUV", 22, "ABCDEFGHIJKLMNOPQRSTUV");
            EndemeConstructor_testcase2("ABCDEFGHIJKLMNOPQRSTUVWXYZ", 26, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
            EndemeConstructor_testcase2("", 0, "");
            EndemeConstructor_testcase2("regf78o$y4trg745hgh(*^uihd", 11, "REGFOYTHUID");

            EndemeConstructor_testcase3("ABCDEFGHIJKLMNOPQRSTUVWXYZ", WetlandAnimals, 22, "ABCDEFGHIJKLMNOPQRSTUV");
            EndemeConstructor_testcase3("", WetlandAnimals, 0, "");
            EndemeConstructor_testcase3("ABCEDA", WetlandAnimals, 5, "ABCED");

            _result = "OK";
        }
        private static void EndemeConstructor_testcase1()
        {
            //EndemeSet set = new EndemeSet();
            Endeme e = new Endeme();
            Assert.That(e.IsEmpty);
        }
        private static void EndemeConstructor_testcase2(string endeme, int length, string target)
        {
            EndemeSet set = new EndemeSet();
            Endeme e = new Endeme(set, endeme);
            int eLength = e.ToCharArray().Length;
            Assert.That(eLength == length);
            string str = e.ToString();
            Assert.That(str == target);
        }
        private static void EndemeConstructor_testcase3(string endeme, EndemeSet set, int length, string target)
        {
            Endeme e = new Endeme(set, endeme);
            int eLength = e.ToCharArray().Length;
            Assert.That(eLength == length);
            string str = e.ToString();
            Assert.That(str == target);
        }


        /// <summary>
        /// 
        /// </summary>
        private void Endeme_Show_test()
        {
            Endeme endeme = new Endeme(WetlandAnimals, "MKACLENSJHUFIGBVTQDPOR");
            string str = "";
            str = endeme.Show_complex(27, 5, "-");            Assert.That(str == "MKA-CLEN-SJHU-FIGB-VTQD-POR"); // perfect, common
            str = endeme.Show_complex(27, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUFIGB-VTQD-POR ");
            str = endeme.Show_complex(27, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFIGB-VTQDPOR  ");
            str = endeme.Show_complex(27, 2, "-");            Assert.That(str == "MKACLEN-SJHUFIGB-VTQDPOR   ");
            str = endeme.Show_complex(27, 1, "-");            Assert.That(str == "MKACLEN-SJHUFIGBVTQDPOR    ");
            str = endeme.Show_complex(27, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR     ");

            str = endeme.Show_complex(26, 5, "-");            Assert.That(str == "MKA-CLEN-SJHUFIG--VTQD-POR");
            str = endeme.Show_complex(26, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUFIGB-VTQD-POR"); // perfect, common
            str = endeme.Show_complex(26, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFIGB-VTQDPOR ");
            str = endeme.Show_complex(26, 2, "-");            Assert.That(str == "MKACLEN-SJHUFIGB-VTQDPOR  ");
            str = endeme.Show_complex(26, 1, "-");            Assert.That(str == "MKACLEN-SJHUFIGBVTQDPOR   ");
            str = endeme.Show_complex(26, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR    ");

            str = endeme.Show_complex(25, 5, "-");            Assert.That(str == "MKA-CLEN-SJHUFI--VTQD-POR");
            str = endeme.Show_complex(25, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUFIG-VTQD-POR");
            str = endeme.Show_complex(25, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFIGB-VTQDPOR"); // perfect
            str = endeme.Show_complex(25, 2, "-");            Assert.That(str == "MKACLEN-SJHUFIGB-VTQDPOR ");
            str = endeme.Show_complex(25, 1, "-");            Assert.That(str == "MKACLEN-SJHUFIGBVTQDPOR  ");
            str = endeme.Show_complex(25, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR   ");

            str = endeme.Show_complex(24, 5, "-");            Assert.That(str == "MKA-CLEN-SJHUFI--TQD-POR");
            str = endeme.Show_complex(24, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUFI-VTQD-POR");
            str = endeme.Show_complex(24, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFIG-VTQDPOR");
            str = endeme.Show_complex(24, 2, "-");            Assert.That(str == "MKACLEN-SJHUFIGB-VTQDPOR"); // perfect, common
            str = endeme.Show_complex(24, 1, "-");            Assert.That(str == "MKACLEN-SJHUFIGBVTQDPOR ");
            str = endeme.Show_complex(24, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR  ");

            str = endeme.Show_complex(23, 5, "-");            Assert.That(str == "MKA-CLEN-SJHUF--TQD-POR");
            str = endeme.Show_complex(23, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUFI-TQD-POR");
            str = endeme.Show_complex(23, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFI-VTQDPOR");
            str = endeme.Show_complex(23, 2, "-");            Assert.That(str == "MKACLEN-SJHUFIG-VTQDPOR");
            str = endeme.Show_complex(23, 1, "-");            Assert.That(str == "MKACLEN-SJHUFIGBVTQDPOR"); // perfect
            str = endeme.Show_complex(23, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR ");

            str = endeme.Show_complex(22, 5, "-");            Assert.That(str == "MKA-CLEN-SJHU--TQD-POR");
            str = endeme.Show_complex(22, 4, "-");            Assert.That(str == "MKA-CLEN-SJHUF-TQD-POR");
            str = endeme.Show_complex(22, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUFI-TQDPOR");
            str = endeme.Show_complex(22, 2, "-");            Assert.That(str == "MKACLEN-SJHUFI-VTQDPOR");
            str = endeme.Show_complex(22, 1, "-");            Assert.That(str == "MKACLENSJHUFIG-VTQDPOR");
            str = endeme.Show_complex(22, 0, "-");            Assert.That(str == "MKACLENSJHUFIGBVTQDPOR"); // perfect

            str = endeme.Show_complex(21, 5, "-");            Assert.That(str == "MKA-CLEN-SJHU--QD-POR");
            str = endeme.Show_complex(21, 4, "-");            Assert.That(str == "MKA-CLEN-SJHU-TQD-POR");
            str = endeme.Show_complex(21, 3, "-");            Assert.That(str == "MKA-CLEN-SJHUF-TQDPOR");
            str = endeme.Show_complex(21, 2, "-");            Assert.That(str == "MKACLEN-SJHUFI-TQDPOR");
            str = endeme.Show_complex(21, 1, "-");            Assert.That(str == "MKACLENSJHUFI-VTQDPOR");
            str = endeme.Show_complex(21, 0, "-");            Assert.That(str == "MKACLENSJHUFIGVTQDPOR");

            str = endeme.Show_complex(20, 5, "-");            Assert.That(str == "MKA-CLEN-SJH--QD-POR");
            str = endeme.Show_complex(20, 4, "-");            Assert.That(str == "MKA-CLEN-SJHU-QD-POR");
            str = endeme.Show_complex(20, 3, "-");            Assert.That(str == "MKA-CLEN-SJHU-TQDPOR");
            str = endeme.Show_complex(20, 2, "-");            Assert.That(str == "MKACLEN-SJHUF-TQDPOR");
            str = endeme.Show_complex(20, 1, "-");            Assert.That(str == "MKACLENSJHUFI-TQDPOR");
            str = endeme.Show_complex(20, 0, "-");            Assert.That(str == "MKACLENSJHUFIVTQDPOR");

            str = endeme.Show_complex(18, 5, "-");            Assert.That(str == "MKA-CLEN-SJ--D-POR");
            str = endeme.Show_complex(18, 4, "-");            Assert.That(str == "MKA-CLEN-SJ-QD-POR");
            str = endeme.Show_complex(18, 3, "-");            Assert.That(str == "MKA-CLEN-SJH-QDPOR");
            str = endeme.Show_complex(18, 2, "-");            Assert.That(str == "MKACLEN-SJHU-QDPOR");
            str = endeme.Show_complex(18, 1, "-");            Assert.That(str == "MKACLENSJHU-TQDPOR");
            str = endeme.Show_complex(18, 0, "-");            Assert.That(str == "MKACLENSJHUFTQDPOR");

            str = endeme.Show_complex(15, 5, "-");            Assert.That(str == "MKA-CLEN----POR");
            str = endeme.Show_complex(15, 4, "-");            Assert.That(str == "MKA-CLEN--D-POR");
            str = endeme.Show_complex(15, 3, "-");            Assert.That(str == "MKA-CLEN-S-DPOR");
            str = endeme.Show_complex(15, 2, "-");            Assert.That(str == "MKACLEN-SJ-DPOR");
            str = endeme.Show_complex(15, 1, "-");            Assert.That(str == "MKACLENSJ-QDPOR");
            str = endeme.Show_complex(15, 0, "-");            Assert.That(str == "MKACLENSJHQDPOR");

            str = endeme.Show_complex(12, 5, "-");            Assert.That(str == "MKA-CL----OR");
            str = endeme.Show_complex(12, 4, "-");            Assert.That(str == "MKA-CL---POR");
            str = endeme.Show_complex(12, 3, "-");            Assert.That(str == "MKA-CLE--POR");
            str = endeme.Show_complex(12, 2, "-");            Assert.That(str == "MKACLEN--POR");
            str = endeme.Show_complex(12, 1, "-");            Assert.That(str == "MKACLEN-DPOR");
            str = endeme.Show_complex(12, 0, "-");            Assert.That(str == "MKACLENSDPOR");

            str = endeme.Show_complex(11, 5, "-");            Assert.That(str == "MKA-C----OR");
            str = endeme.Show_complex(11, 4, "-");            Assert.That(str == "MKA-CL---OR");
            str = endeme.Show_complex(11, 3, "-");            Assert.That(str == "MKA-CL--POR");
            str = endeme.Show_complex(11, 2, "-");            Assert.That(str == "MKACLE--POR");
            str = endeme.Show_complex(11, 1, "-");            Assert.That(str == "MKACLEN-POR"); // common
            str = endeme.Show_complex(11, 0, "-");            Assert.That(str == "MKACLENDPOR");

            str = endeme.Show_complex(10, 5, "-");            Assert.That(str == "MKA-----OR");
            str = endeme.Show_complex(10, 4, "-");            Assert.That(str == "MKA-C---OR");
            str = endeme.Show_complex(10, 3, "-");            Assert.That(str == "MKA-CL--OR");
            str = endeme.Show_complex(10, 2, "-");            Assert.That(str == "MKACL--POR");
            str = endeme.Show_complex(10, 1, "-");            Assert.That(str == "MKACLE-POR");
            str = endeme.Show_complex(10, 0, "-");            Assert.That(str == "MKACLENPOR");

            // phase change here from a 33% low, 67% high approach to 100% high

            str = endeme.Show_complex( 8, 5, "-");            Assert.That(str == "MKA-----");
            str = endeme.Show_complex( 8, 4, "-");            Assert.That(str == "MKA-C---");
            str = endeme.Show_complex( 8, 3, "-");            Assert.That(str == "MKA-CL--");
            str = endeme.Show_complex( 8, 2, "-");            Assert.That(str == "MKA-CLE-");
            str = endeme.Show_complex( 8, 1, "-");            Assert.That(str == "MKA-CLEN"); // common
            str = endeme.Show_complex( 8, 0, "-");            Assert.That(str == "MKACLENS");

            str = endeme.Show_complex( 7, 5, "-");            Assert.That(str == "MK-----");
            str = endeme.Show_complex( 7, 4, "-");            Assert.That(str == "MKA----");
            str = endeme.Show_complex( 7, 3, "-");            Assert.That(str == "MKA-C--");
            str = endeme.Show_complex( 7, 2, "-");            Assert.That(str == "MKA-CL-");
            str = endeme.Show_complex( 7, 1, "-");            Assert.That(str == "MKA-CLE");
            str = endeme.Show_complex( 7, 0, "-");            Assert.That(str == "MKACLEN"); // common

            str = endeme.Show_complex( 5, 5, "-");            Assert.That(str == "-----");
            str = endeme.Show_complex( 5, 4, "-");            Assert.That(str == "M----");
            str = endeme.Show_complex( 5, 3, "-");            Assert.That(str == "MK---");
            str = endeme.Show_complex( 5, 2, "-");            Assert.That(str == "MKA--");
            str = endeme.Show_complex( 5, 1, "-");            Assert.That(str == "MKA-C");
            str = endeme.Show_complex( 5, 0, "-");            Assert.That(str == "MKACL");

            str = endeme.Show_complex( 3, 5, "-");            Assert.That(str == "---");
            str = endeme.Show_complex( 3, 4, "-");            Assert.That(str == "---");
            str = endeme.Show_complex( 3, 3, "-");            Assert.That(str == "---");
            str = endeme.Show_complex( 3, 2, "-");            Assert.That(str == "M--");
            str = endeme.Show_complex( 3, 1, "-");            Assert.That(str == "MK-");
            str = endeme.Show_complex( 3, 0, "-");            Assert.That(str == "MKA"); // common

            str = endeme.Show_complex( 1, 5, "-");            Assert.That(str == "-");
            str = endeme.Show_complex( 1, 4, "-");            Assert.That(str == "-");
            str = endeme.Show_complex( 1, 3, "-");            Assert.That(str == "-");
            str = endeme.Show_complex( 1, 2, "-");            Assert.That(str == "-");
            str = endeme.Show_complex( 1, 1, "-");            Assert.That(str == "-");
            str = endeme.Show_complex( 1, 0, "-");            Assert.That(str == "M");

            str = endeme.Show_complex( 0, 5, "-");            Assert.That(str == "");
            str = endeme.Show_complex( 0, 4, "-");            Assert.That(str == "");
            str = endeme.Show_complex( 0, 3, "-");            Assert.That(str == "");
            str = endeme.Show_complex( 0, 2, "-");            Assert.That(str == "");
            str = endeme.Show_complex( 0, 1, "-");            Assert.That(str == "");
            str = endeme.Show_complex( 0, 0, "-");            Assert.That(str == "");

            str = endeme.Show_complex(-1, 5, "-");            Assert.That(str == "");
            str = endeme.Show_complex(-1, 4, "-");            Assert.That(str == "");
            str = endeme.Show_complex(-1, 3, "-");            Assert.That(str == "");
            str = endeme.Show_complex(-1, 2, "-");            Assert.That(str == "");
            str = endeme.Show_complex(-1, 1, "-");            Assert.That(str == "");
            str = endeme.Show_complex(-1, 0, "-");            Assert.That(str == "");
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeCharacteristic_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        private void EndemeCharacteristic_Constructor_test()
        {
            EndemeCharacteristic ec = new EndemeCharacteristic('C', "Cat", "Felines");
            Assert.That(ec.Label == "Cat");
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeDistribution_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void EndemeDistribution_test()
        {
            _result = "";
            int count = 512;
            Endeme[] listA = new Endeme[count];
            Endeme[] listB = new Endeme[count];
            EndemeSet animals = WetlandAnimals;
            Random r = new Random();


            // --------------------------------------------------------------------------
            //  Create a bunch of random endemes
            // --------------------------------------------------------------------------
            DateTime stamp1 = DateTime.Now;
            for (int j = 0; j < count; ++j)
            {
                listA[j] = animals.RandomEndeme(r);
                listB[j] = animals.RandomEndeme(r);
            }
            DateTime stamp2 = DateTime.Now;
            TimeSpan operation1 = stamp2 - stamp1;
            _result += "Creating "+count*2+" endemes took " + operation1.TotalSeconds + " seconds.";


            // --------------------------------------------------------------------------
            //  Match the endemes and EndemeTestTally
            // --------------------------------------------------------------------------
            EndemeTestTally tally = new EndemeTestTally(100);
            for (int i = 0; i < count; ++i)
            {
                for (int j = 0; j < count; ++j)
                {
                    Endeme e1 = listA[i];
                    Endeme e2 = listB[j];
                    double match = animals.Match(e1, e2);
                    Assert.That(0.0 < match && match < 1.0 || e1 == e2 || e1 == e2.Reverse);
                    tally.Incr((int)(match * 99.0 + 1.0), 2);
                }
            }
            DateTime stamp3 = DateTime.Now;
            TimeSpan operation2 = stamp3 - stamp2;
            string report = tally.Report("" + count*count + " endeme matches"
                + " took " + operation2.TotalSeconds + " seconds", 60);
            _result += "\r\n" + report;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeLeaf_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        private void EndemeLeaf_Constructor_test()
        {
            Endeme endeme = WetlandAnimals.RandomEndeme();
            EndemeItem item = new EndemeItem(endeme);
            item.Label = "hi";
            item.Number = 7;
            item.Text = "fosdfjej";

            Assert.That(item.Endeme.ToString() == endeme.ToString());
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeList_Constr_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void EndemeList_Constructor_test()
        {
            //_result = "";
            //EndemeSet animals = WetlandAnimals;
            //EndemeList list = new EndemeList(100, animals, true);
            //Random r = EndemeRandomSource.New().Random;
            //for (int i = 0; i < 100; i++)
            //{
            //    list.Add(Guid.Empty, i.ToString(), animals.RandomEndeme(r));
            //}
            //List<Guid> key = list["47"];
            //Endeme     e   = list[key[0]];
            //string     str = e.ToString();
            //Assert.That(str.Length == 22);

            //_result = "OK";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeMatch_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void Endeme_Match_test()
        {
            _result = "";
            EndemeSet animals = WetlandAnimals;
            Random r = new Random();


            Endeme_Match_testcase1(animals, "ABCDEFGHIJKLMNOPQRSTUV", "VUTSRQPONMLKJIHGFEDCBA", 0.0);
            Endeme_Match_testcase1(animals, "ABCDEFGHIJKLMNOPQRSTUV", "ABCDEFGHIJKLMNOPQRSTUV", 1.0);
            Endeme_Match_testcase2(r, animals);

            for (int i = -20; i < 100; ++i)
            {
                Endeme_Match_testcase1(animals, "ABCDEFGHIJKLMNOPQRSTUV", "VUTSRQPONMLKJIHGFEDCBA", 0.0);
                Endeme_Match_testcase1(animals, "ABCDEFGHIJKLMNOPQRSTUV", "ABCDEFGHIJKLMNOPQRSTUV", 1.0);
                Endeme_Match_testcase2(r, animals);
            }
        }
        private static void Endeme_Match_testcase1(EndemeSet set, string endeme1, string endeme2
            , double target)
        {
            Endeme e1 = new Endeme(set, endeme1);
            Endeme e2 = new Endeme(set, endeme2);
            double match = set.Match(e1, e2);
            Assert.That(match == target);
        }
        private static double Endeme_Match_testcase2(Random r, EndemeSet set)
        {
            Endeme e1 = set.RandomEndeme(r);
            Endeme e2 = set.RandomEndeme(r);
            double match = set.Match(e1, e2);
            Assert.That(0.0 < match && match < 1.0 || e1 == e2 || e1 == e2.Reverse);
            return match;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Endeme_MatchOrder_test -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="animals"></param>
        //[Test]
        public void Endeme_MatchOrder_test()
        {
            EndemeSet animals = WetlandAnimals;
            Endeme e1 = new Endeme(animals, "ABCDEFGHIJKLMNOPQRSTUV");
            Endeme e2 = new Endeme(animals, "DEFGHIJKLMNOCPQRSTUVAB"); // difference = 100
            double match2 = animals.Match(e1, e2); int diff2 = animals.Diff(e1, e2);
            Endeme e3 = new Endeme(animals, "ABCDRQPONMLKJIHGFSETUV"); // difference = 100
            double match3 = animals.Match(e1, e3); int diff3 = animals.Diff(e1, e3);
            Endeme e4 = new Endeme(animals, "JIHGFEDCBAKLVUTSRQPONM"); // difference = 100
            double match4 = animals.Match(e1, e4); int diff4 = animals.Diff(e1, e4);
            Assert.That(diff2 == diff3 && diff3 == diff4);
            Assert.That(match2 < match3 && match3 < match4);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeNormalizer_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        private void EndemeNormalizer_Constructor_test()
        {
            Endeme avg = WetlandAnimals.RandomEndeme();
            Endeme std = WetlandAnimals.RandomEndeme();
            EndemeBalancer norm = new EndemeBalancer(100.0, 10.0, avg, std);

            Assert.That(norm.StdTgt == 10.0);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeSet_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        public void EndemeSet_Constructor_test()
        {
            EndemeSet animals = WetlandAnimals;
            Assert.That(animals.Count == 22);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeSet_ToString_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void EndemeSet_ToString_test()
        {
            _result = "";
            EndemeSet animals = WetlandAnimals;
            string str = animals.ToString();
            Assert.That(str.Length > 20);
            Assert.That(str == "A)alligator, B)beaver, C)crocodile, D)duck, E)egret, F)frog, G)gecko, H)herron, I)insect, J)jackal, K)koala, L)lizard, M)muskrat, N)newt, O)otter, P)puma, Q)quahog, R)reptile, S)snake, T)turtle, U)ungulate, V)vole");

            _result = "OK";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IndexComponent_Constructor_test -->
        /// <summary>
        /// 
        /// </summary>
        //[Test]
        public void IndexComponent_Constructor_test()
        {
            //EndemeList list = new EndemeList(10, WetlandAnimals, false);
            ////EndemeIndexComponent c = new EndemeIndexComponent(list, list.Set.RandomEndeme());
            //Assert.That(c.KeyEndeme == null);
            //Assert.That(c.Count == 0);


            //EndemeList list2 = new EndemeList(10, WetlandAnimals, true);
            //EndemeIndexComponent c2 = new EndemeIndexComponent(list2, list2.Set.RandomEndeme());
            //Assert.That(c2.KeyEndeme.Count == 22);
            //Assert.That(c2.Count == 0);


            //_result = "OK";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- WetlandAnimals -->
        /// <summary>
        ///      Test endeme set
        /// </summary>
        private static EndemeSet WetlandAnimals { get
        {
            EndemeSet set = new EndemeSet("animal");
            set.Add('A', "alligator", "");
            set.Add('B', "beaver",    "");
            set.Add('C', "crocodile", "");
            set.Add('D', "duck",      "");
            set.Add('E', "egret",     "");
            set.Add('F', "frog",      "");
            set.Add('G', "gecko",     "");
            set.Add('H', "herron",    "");
            set.Add('I', "insect",    "");
            set.Add('J', "jackal",    "");
            set.Add('K', "koala",     "");
            set.Add('L', "lizard",    "");
            set.Add('M', "muskrat",   "");
            set.Add('N', "newt",      "");
            set.Add('O', "otter",     "");
            set.Add('P', "puma",      "");
            set.Add('Q', "quahog",    "");
            set.Add('R', "reptile",   "");
            set.Add('S', "snake",     "");
            set.Add('T', "turtle",    "");
            set.Add('U', "ungulate",  "");
            set.Add('V', "vole",      "");
            return set;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Concatenates two strings if neither is blank
        /// </summary>
        /// <remarks>
        ///      This is copied here to keep the namespace atomic
        /// </remarks>
        /// <param name="str1"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static string Affix(string str1, string str2)
        {
            if (!string.IsNullOrEmpty(str1)
                && !string.IsNullOrEmpty(str2))
                return str1 + str2;
            else
                return "";
        }


    }


    // --------------------------------------------------------------------------------------------
    /// <!-- Assert -->
    /// <summary>
    /// 
    /// </summary>
    internal static class Assert
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="ok"></param>
        /// <returns></returns>
        public static bool That(bool ok)
        {
            if (!ok) throw new Exception("test mismatch");
            else return true;
        }
    }


    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeTestTally -->
    /// <summary>
    /// 
    /// </summary>
    public class EndemeTestTally : Dictionary<string,int>
    {
        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        public EndemeTestTally() { }
        public EndemeTestTally(int capacity) : base(capacity) { }


        // ----------------------------------------------------------------------------------------
        /// <!-- Average -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public double Average()
        {
            return ((double)TotalValue() / (double)TotalCount);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Histogram -->
        /// <summary>
        ///      Returns an ASCII art histogram
        /// </summary>
        /// <returns></returns>
        public string Histogram(double divisor)
        {
            StringBuilder str = new StringBuilder();
            List<string> keys = Keys();
            string delim = "";
            int num;


            int width = (int)(this[Mode()] / divisor) + 1;
            int runningCount = 0;


            // -------------------------------------------------------------------------90
            //  Turn the EndemeTestTally list into a histogram
            // -------------------------------------------------------------------------90
            foreach (string key in keys)
            {
                int count = this[key];
                num = (int)((1.0 * count) / divisor + 0.5);
                str.Append(delim + key + "\t");
                str.Append(Stars(num, '*'));
                str.Append(Stars(width - num, ' '));
                str.Append("\t" + count);
                runningCount += count;
                str.Append("\t" + runningCount);
                delim = "\r\n";
            }


            return str.ToString();
        }


        public string Histogram()          { return Histogram(1.0); }
        public string Histogram(int width) { return Histogram((double)MaxTally/(double)width); }


        // ----------------------------------------------------------------------------------------
        /// <!-- Incr -->
        /// <summary>
        ///      Increments one item in the EndemeTestTally,
        ///      adds the item if it is not there yet starting it with 1
        /// </summary>
        /// <param name="count"></param>
        /// <param name="msg"></param>
        public void Incr(string value)
        {
            if (ContainsKey(value))
                this[value]++;
            else Add(value, 1);
        }
        public void Incr(string value, bool includeBlanks)
        {
            if (includeBlanks || !string.IsNullOrEmpty(value))
                { if (ContainsKey(value)) this[value]++; else Add(value, 1); }
        }
        /// <summary>
        ///      Stores zero padded numbers as string labels so they can be sorted alphabetically,
        ///      if you don't want zero padding, turn them into strings first
        /// </summary>
        /// <param name="label"></param>
        /// <param name="zeroPaddingWidth">the width of the number when stored as a string and when displayed</param>
        public void Incr(int value, int zeroPaddingWidth)
        {
            Incr(value.ToString().PadLeft(zeroPaddingWidth, '0'));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Keys -->
        /// <summary>
        ///      Returns a sorted list of keys (overrides Keys), sorts alphabetically, or
        ///      numerically if you put them in with enough zero padding
        /// </summary>
        /// <returns></returns>
        private new List<string> Keys()
        {
            List<string> list = new List<string>(this.Count);
            foreach (string key in base.Keys)
                list.Add(key);
            list.Sort();
            return list;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Max -->
        /// <summary>
        ///      Returns the manimum (or final) key, works whether the numbers are labels or strings
        /// </summary>
        /// <returns></returns>
        public string Max()
        {
            List<string> key = Keys();
            return key[key.Count-1];
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- MaxTally -->
        /// <summary>
        ///      Returns the maximum EndemeTestTally
        /// </summary>
        /// <returns></returns>
        public int MaxTally { get { return this[Mode()]; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Median -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string Median()
        {
            List<string> keys = Keys();
            int halfWay = TotalCount / 2;
            string median = "";


            int runningTalley = 0;
            foreach (string key in keys)
            {
                runningTalley += this[key];
                if (runningTalley > halfWay)  { median = key;  break; }
            }
            return median;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Min -->
        /// <summary>
        ///      Returns the minimum (or first) key, works whether the numbers are labels or strings
        /// </summary>
        /// <returns></returns>
        public string Min()
        {
            List<string> key = Keys();
            if (key.Count > 0)  return key[0];
            else return "";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Mode -->
        /// <summary>
        ///      This works whether the labels are numbers or strings
        /// </summary>
        /// <returns></returns>
        public string Mode()
        {
            List<string> keys = Keys();


            string mode = "";
            int maxCount = 0;
            foreach (string key in keys)
                if (this[key] > maxCount)
                    { maxCount = this[key];  mode = key; }
            return mode;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Report -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="reportTitle"></param>
        /// <returns></returns>
        public string Report(string reportTitle, int width)
        {
            try
            {
                StringBuilder str = new StringBuilder(reportTitle);
                str.Append("\r\n");
                str.Append("\r\n" + ToString());
                str.Append("\r\n");
                double total = TotalValue(); str.Append("\r\nTotal:   " + total);
                int    count = TotalCount;   str.Append("\r\nCount:   " + count);
                int    items = Count;        str.Append("\r\nItems:   " + items);
                string min   = Min();        str.Append("\r\nMin:     " + min  + " (" + this[min]  + ")");
                double mean  = Average();    str.Append("\r\nMean:    " + mean);
                string med   = Median();     str.Append("\r\nMedian:  " + med  + " (" + this[med]  + ")");
                string mode  = Mode();       str.Append("\r\nMode:    " + mode + " (" + this[mode] + ")");
                string max   = Max();        str.Append("\r\nMax:     " + max  + " (" + this[max]  + ")");
                str.Append("\r\n");
                str.Append("\r\n" + Histogram(width));
                return str.ToString();
            }
            catch (Exception ex)
            {
                return "EndemeTestTally.Report encountered an error - " + ex.Message;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Stars -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="num"></param>
        /// <param name="cha"></param>
        /// <returns></returns>
        public static string Stars(int num, char cha)
        {
            string stars = "";
            for (int i = 0; i < num; ++i)
                stars += cha;
            return stars;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        ///      Turns the EndemeTestTally list into a comma and colon delimited string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder(Count);
            string delim = "";


            // --------------------------------------------------------------------------
            //  Do stuff
            // --------------------------------------------------------------------------
            List<string> list = Keys();
            foreach (string key in list)
            {
                str.Append(delim + key + ":" + this[key]);
                delim = ", ";
            }


            return str.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- TotalCount -->
        /// <summary>
        ///      Returns the total EndemeTestTally by adding up all the sub-tallys, works whether the numbers
        ///      are labels or strings
        /// </summary>
        /// <returns></returns>
        public int TotalCount
        {
            get
            {
                List<string> keys = Keys();
                int total = 0;
                foreach (string key in keys)
                    total += this[key];
                return total;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- TotalValue -->
        /// <summary>
        ///      Returns the total EndemeTestTally by adding up all the sub-tallys, only works if the labels
        ///      are numbers
        /// </summary>
        /// <returns></returns>
        public int TotalValue()
        {
            List<string> keys = Keys();
            int total = 0;
            foreach (string key in keys)
            {
                int val = 0;
                int value;
                if (int.TryParse(key, out value))
                    val = value;
                total += val * this[key];
            }
            return total;
        }



    }

}
