﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeUIField -->
    /// <summary>
    ///      The EndemeUIField class contains UI parameters for a data field whose context is defined
    ///      by an endeme
    /// </summary>
    public class EndemeUIField
    {
        public EndemeSet Set       { get; set; }

        public string    Label     { get; set; }
        public int       Padding   { get; set; } // width
        public int       Precision { get; set; }
        public char      ToDisplay { get; set; } // E:the endeme itself, N:number, R:Raw, T:text, V:value
        public string    Header    { get; set; }


        // ----------------------------------------------------------------------------------------
        /// <!-- EndemeUIField constructor -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="header"></param>
        /// <param name="label"></param>
        /// <param name="padding"></param>
        /// <param name="toShow"></param>
        /// <param name="endemeSet"></param>
        public EndemeUIField(string header, string label, int padding, char toShow, EndemeSet endemeSet)
        {
            Header    = header;
            Label     = label;
            Padding   = padding;
            Precision = 0;
            ToDisplay = toShow;
            Set       = endemeSet;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowData -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public string ShowData(EndemeItem item)
        {
            string str = "";
            switch (ToDisplay)
            {
                case 'E': str = item.Endeme.Show_complex(Padding, 4, "-");              break;
                case 'N': str = ((int)item.Number).ToString().PadLeft(Padding); break;
                case 'R': foreach (char cha in Set.Letters) str += ((int)item.Endeme.Quant.Raw[cha]).ToString().PadLeft(3);   break;
                case '|': str = this.Label.PadLeft(Padding);                    break;
            }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowHeader -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string ShowHeader()
        {
            string str = "";
            switch (ToDisplay)
            {
                case 'E': str = Header.PadLeft(Padding); break;
                case 'N': str = Header.PadLeft(Padding); break;
                case 'R': foreach (char letter in Set.Letters) str += letter.ToString().PadLeft(Padding); break;
                case '|': str = Header.PadLeft(Padding); break;
            }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowLiteral -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public string ShowLiteral(double data, Endeme endeme)
        {
            if (endeme == null)
                return " ".PadLeft(Padding);
            string str = "";
            switch (ToDisplay)
            {
                case 'E': str = endeme.Show_complex(Padding, 4, "-");             break;
                case 'N': str = ((int)data).ToString().PadLeft(Padding); break;
                case 'R': foreach (char cha in Set.Letters) str += ((int)endeme.Quant.Raw[cha]).ToString().PadLeft(Padding); break;
                case '|': str = this.Label.PadLeft(Padding);             break;
            }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowRowDivider -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string ShowRowDivider()
        {
            string str = "";
            switch (ToDisplay)
            {
                case 'E': str = "-".PadRight(Padding, '-'); break;
                case 'N': str = " ".PadRight(Padding, '-'); break;
                case 'R': foreach (char c in Set.Letters) str += " ".PadRight(Padding, '-'); break;
                case '|': str = "+".PadLeft(Padding); break;
            }
            return str;
        }
    }
}
