﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;

namespace InformationLib.SoftData // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- Roll -->
    /// <summary>
    ///      The Roll class contains a bunch of functions many of which have to do with generating random things
    /// </summary>
    public static class Roll
    {
        private static Random _r;


        // ----------------------------------------------------------------------------------------
        //  Short methods
        // ----------------------------------------------------------------------------------------
        public  static int  D2      { get { Init(); return 1 + _r.Next(2);                        } }
        public  static int  D6      { get { Init(); return 1 + _r.Next(6);                        } }
        public  static int  _4D6    { get { return D6 + D6 + D6 + D6;                             } }
        public  static int  OneD  (int n) { Init(); return _r.Next(n) + 1;                        }
        public  static int  TwoD  (int n) { return OneD(n) + Plus1D(n);                           }
        public  static int  ThreeD(int n) { return OneD(n) + Plus1D(n) + Plus1D(n);               }
        public  static int  FourD (int n) { return OneD(n) + Plus3D(n);                           }
        public  static int  FiveD (int n) { return OneD(n) + Plus4D(n);                           }
        public  static int  SixD  (int n) { return OneD(n) + Plus4D(n) + Plus1D(n);               }
        public  static int  SevenD(int n) { return OneD(n) + Plus4D(n) + Plus1D(n) + Plus1D(n);   }
        // ----------------------------------------------------------------------------------------
        private static int  Plus1D(int n) { return _r.Next(n) + 1;                                }
        private static int  Plus3D(int n) { return Plus1D(n) + Plus1D(n) + Plus1D(n);             }
        private static int  Plus4D(int n) { return Plus3D(n) + Plus1D(n);                         }
        private static void Init  ()      { _r = DataRandomSource.New().Random;                   }


        // ----------------------------------------------------------------------------------------
        /// <!-- Jiggle -->
        /// <summary>
        ///      Jiggle a number by 1 within the specified limits
        /// </summary>
        /// <remarks>
        ///      This may belong in the RandomSource class
        /// </remarks>
        /// <param name="r"></param>
        /// <param name="min">lower inclusive limit</param>
        /// <param name="curr">the current value of the number to be modified</param>
        /// <param name="max">higher inclusive limit</param>
        /// <returns></returns>
        public static int Jiggle(Random r, int min, int curr, int max)
        {
            curr = curr + r.Next(3) - 1;
            if (curr < min) curr = min;
            if (curr > max) curr = max;
            return curr;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Max -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <param name="d"></param>
        /// <returns></returns>
        public static double Max(double a, double b, double c, double d)
        {
            double max = Math.Max(a, b);
            max = Math.Max(max, c);
            max = Math.Max(max, d);
            return max;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Nd6 -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="n"></param>
        /// <returns></returns>
        public static int Nd6(int n)
        {
            _r = DataRandomSource.New().Random;
            int sum = 0;
            for (int i = 0; i < n; ++i)
                sum += D6;
            return sum;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Point -->
        /// <summary>
        ///      Returns a random point
        /// </summary>
        /// <param name="X_hi"></param>
        /// <param name="Y_hi"></param>
        /// <returns></returns>
        public static Point Point(int X_hi, int Y_hi)
        {
            return new Point(_r.Next(X_hi), _r.Next(Y_hi));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StdDeviation -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="list"></param>
        /// <returns></returns>
        public static double StdDeviation(Dictionary<char, int> list)
        {
            double average = list.Values.Average();
            double sumOfDifferencesSquared = list.Select(val => ((double)val.Value - average) * ((double)val.Value - average)).Sum();
            double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
            return sd;
        }


        public static double StdDeviation(List<double> list)
        {
            double average = list.Average();
            double sumOfDifferencesSquared = list.Select(val => ((double)val - average) * ((double)val - average)).Sum();
            double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
            return sd;
        }
    }
}
