﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Data;                                // for DataException
using System.Data.SqlTypes;                       // for SqlDateTime
using System.IO;                                  // for File, StreamWriter, TextReader
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;             // for Regex
using System.Xml;                                 // for XmlNode

namespace InformationLib.SoftData // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- TreatAs -->
    /// <summary>
    /// 
    /// </summary>
    public static class TreatAs
    {

        // ----------------------------------------------------------------------------------------
        /// <!-- BoolValue -->
        /// <summary>
        ///      Takes its best shot at converting whatever is sent to it to a boolean value
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue">returns this for blank, null, or unidentified</param>
        /// <returns></returns>
        public static bool BoolValue(object obj, bool defaultValue)
        {
            if (IsNull(obj))
            {
                return defaultValue;
            }
            else
            {
                Type type = obj.GetType();
                string str;
                //bool value;


                // ----------------------------------------------------------------------
                //  Standard approaches
                // ----------------------------------------------------------------------
                if (type == typeof(bool) || type == typeof(Boolean))
                    return (bool)obj;
                str = obj.ToString();
                //if (bool.TryParse(str, out value))  return value;


                // ----------------------------------------------------------------------
                //  Blank values
                // ----------------------------------------------------------------------
                //if (str == "") return defaultValue;
                if (str == type.ToString())
                    throw new DataException("_.Bool - ToString() not defined in any useful way for " + type.ToString());


                return ((bool)SqlBool(str, defaultValue));
                //// --------------------------------------------------------------------
                ////  Additional approaches - 
                ////  1/0, acceptable/unacceptable, complete/incomplete, correct/incorrect
                ////  , engaged/disengaged, equal/unequal, good/bad, high/low, in/out
                ////  , non-null/null, ok/not ok, open/closed, positive/negative
                ////  , present/absent, right/wrong, true/false, yes/no
                ////  etc.
                //// --------------------------------------------------------------------
                //if (Regex.IsMatch(str, "^(acc|com|cor|non-nul|nonnul)",   RegexOptions.IgnoreCase)) return true;
                //if (Regex.IsMatch(str, "^(abs|closed|inco|neg|not ok)",   RegexOptions.IgnoreCase)) return false;
                //if (Regex.IsMatch(str, "^(good|hi|high|in|ok|open|yes)$", RegexOptions.IgnoreCase)) return true;
                //if (Regex.IsMatch(str, "^(bad|lo|low|no|nil|null?|out)$", RegexOptions.IgnoreCase)) return false;


                //// --------------------------------------------------------------------
                ////  letters and symbols:
                ////  Y/N, T/F, R/W, A/U, E/D, G/B, H/L, C/I, JMOPV/KQSXZ, +/-, &/|, 1-9/0, >/<
                //// --------------------------------------------------------------------
                //if (Regex.IsMatch(str, "^[YTRAEGHCJMOPV]", RegexOptions.IgnoreCase)) return true;  // AbCd EfGH iJklMn OPqRsT uVwxYz
                //if (Regex.IsMatch(str, "^[NFWUDBLIKQSXZ]", RegexOptions.IgnoreCase)) return false;
                //if (Regex.IsMatch(str, "^[+1-9&>]",        RegexOptions.IgnoreCase)) return true;
                //if (Regex.IsMatch(str, "^[-0|<]",          RegexOptions.IgnoreCase)) return false;


                //return defaultValue;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Clip -->
        /// <summary>
        ///       Constrains an integer to a specific range
        /// </summary>
        /// <param name="num"></param>
        /// <param name="lowerLimit"></param>
        /// <param name="upperLimit"></param>
        /// <returns></returns>
        public static int Clip(int num, int lowerLimit, int upperLimit)
        {
            return Math.Min(Math.Max(num, lowerLimit), upperLimit);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- DecValue -->
        /// <summary>
        ///      Takes a shot at converting what is sent to it to a decimal
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static decimal DecValue(object obj, decimal defaultValue)
        {
            if (IsNull(obj))
                return defaultValue;
            else
            {
                Type type = obj.GetType();
                string str;
                decimal value;


                // ----------------------------------------------------------------------
                //  Standard approaches
                // ----------------------------------------------------------------------
                if (type == typeof(decimal) || type == typeof(Decimal))
                    return (decimal)obj;
                str = obj.ToString();
                if (decimal.TryParse(str, out value)) return value;


                return defaultValue;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- FloatValue -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="obj">object to be cast as a float</param>
        /// <param name="defaultValue">value to be returned if you can't cast</param>
        /// <returns></returns>
        public static float FloatValue(object obj, float defaultValue)
        {
            if (IsNull(obj))
                return defaultValue;
            else
            {
                Type type = obj.GetType();
                string str;
                float value;


                // ----------------------------------------------------------------------
                //  Standard approaches
                // ----------------------------------------------------------------------
                if (type == typeof(float)) return (float)obj;
                if (type == typeof(double) || type == typeof(Double))
                {
                    double num = (double)obj;
                    if (float.MinValue <= num && num <= float.MaxValue) return (float)num;
                    else return defaultValue;
                }
                str = obj.ToString();
                if (float.TryParse(str, out value)) return value;


                return defaultValue;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- GuidValue -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static Guid GuidValue(object obj, Guid defaultValue)
        {
            Guid g = new Guid();
            if (!IsNull(obj))
            {
                if (obj.GetType() == typeof(Guid)) return (Guid)obj;


                // ----------------------------------------------------------------------
                //  Try various string formats
                // ----------------------------------------------------------------------
                string str = obj.ToString();
                if (str.Length == 22 || str.Length == 24)
                    try { g = GUID(str); return g; }
                    catch { }
                else
                    if (IsGuid(str, out g)) return g;


                g = defaultValue;
            }
            else
                g = defaultValue;


            return defaultValue;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- GUID -->
        /// <summary>
        ///      Convert base64 string format to GUID (From Dave Transom's C# Vitamins)
        /// </summary>
        /// <param name="strGuid64"></param>
        /// <returns></returns>
        public static Guid GUID(string strGuid64)
        {
            string value = strGuid64.Replace("_", "/").Replace("-", "+");
            byte[] buffer = Convert.FromBase64String(value + "==");
            Guid guid = new Guid(buffer);
            return guid;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- GUID -->
        /// <summary>
        ///      Converts a GUID to a base 64 string (From Dave Transom's C# Vitamins)
        /// </summary>
        /// <param name="guid"></param>
        /// <returns></returns>
        public static string GUID(Guid guid, string defaultValue)
        {
            string str64;
            try
            {
                string encoded = Convert.ToBase64String(guid.ToByteArray());
                str64 = encoded.Replace("/", "_").Replace("+", "-").Substring(0, 22);
            }
            catch { str64 = defaultValue; }
            return str64;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Guid_0123456789abcdefx2 -->
        /// <summary>
        ///      {01234567-89ab-cdef-0123-456789abcdef}
        /// </summary>
        public static Guid Guid_0123456789abcdefx2
        { get { Guid g = new Guid("01234567-89ab-cdef-0123-456789abcdef"); return g; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Guid_Guid_01234567x4 -->
        /// <summary>
        ///      {00001111-2222-3333-4444-555566667777}
        /// </summary>
        public static Guid Guid_01234567x4
        { get { Guid g = new Guid("00001111-2222-3333-4444-555566667777"); return g; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- GuidChange -->
        /// <summary>
        ///      Did the Guid change and did the conversion succeed?
        /// </summary>
        /// <param name="initialID"></param>
        /// <param name="finalID"></param>
        /// <returns>1:yes,changed  0:no,did not changed  -1:conversion failed,final guid empty</returns>
        public static int GuidChange(Guid initialID, Guid finalID)
        {
            int val;
            if (finalID.Equals(Guid.Empty)) val = -1;
            else if (finalID.Equals(initialID)) val = 0;
            else val = 1;
            return val;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- HighBit -->
        /// <summary>
        ///      Returns the position of the highest bit or -1 for a zero
        /// </summary>
        /// <param name="num"></param>
        /// <returns></returns>
        public static int HighBit(ulong num)
        {
            int count = -1;
            while (num > 0) { count++; num >>= 1; }
            return count;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IntValue -->
        /// <summary>
        ///       Returns an integer from a string within a specific range with a specific default
        /// </summary>
        /// <param name="str"></param>
        /// <param name="defaultNum"></param>
        /// <param name="lowerLimit"></param>
        /// <returns></returns>
        public static int IntValue(string strNum, int lowerLimit, int defaultNum, int upperLimit)
        {
            int num;
            int result = defaultNum;
            if (int.TryParse(strNum, out num))
                result = Clip(num, lowerLimit, upperLimit);
            return result;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IntValue -->
        /// <summary>
        ///      Takes a shot at converting what is sent to it to an integer
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static int IntValue(object obj, int defaultValue)
        {
            if (IsNull(obj))
                return defaultValue;
            else
            {
                Type type = obj.GetType();
                string str;
                int value;


                // ----------------------------------------------------------------------
                //  Standard approaches
                // ----------------------------------------------------------------------
                if (type == typeof(int) || type == typeof(Int16) || type == typeof(Int32)
                    || type == typeof(Int64))
                    return (int)obj;
                str = obj.ToString();
                if (int.TryParse(str, out value)) return value;


                return defaultValue;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IsGuid -->
        private static Regex _isGuid = new Regex(@"^(\{){0,1}[0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}(\}){0,1}$", RegexOptions.Compiled);
        /// <summary>
        ///      VIA Scott Galloway's Blog - In this implementation you pass in a Guid as an 'out'
        ///      parameter along with the string you want to test - it then fills in the Guid and
        ///      returns true / false depending on whether the Guid was valid... 
        /// </summary>
        /// <param name="candidate"></param>
        /// <param name="output"></param>
        /// <returns></returns>
        internal static bool IsGuid(string candidate, out Guid output)
        {
            bool isValid = false;
            output = Guid.Empty;
            if (candidate != null)
            {
                if (_isGuid.IsMatch(candidate))
                {
                    output = new Guid(candidate);
                    isValid = true;
                }
            }
            return isValid;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IsNull -->
        /// <summary>
        ///      Determines whether an object is one or another of the standard null values
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public static bool IsNull(object obj)
        {
            // --------------------------------------------------------------------------
            //  Check for standard null values
            // --------------------------------------------------------------------------
            if (obj == null || obj == DBNull.Value)
                return true;


            // --------------------------------------------------------------------------
            //  Investigate the specific Sql types for null value
            // --------------------------------------------------------------------------
            string type = obj.GetType().ToString();
            if (Regex.IsMatch(type, "^System.Data.SqlTypes"))
                switch (type)
                {
                    case "System.Data.SqlTypes.SqlBoolean" : if (((SqlBoolean)obj).IsNull)  return true; break;
                    case "System.Data.SqlTypes.SqlByte"    : if (((SqlByte)obj).IsNull)     return true; break;
                    case "System.Data.SqlTypes.SqlDateTime": if (((SqlDateTime)obj).IsNull) return true; break;
                    case "System.Data.SqlTypes.SqlDecimal" : if (((SqlDecimal)obj).IsNull)  return true; break;
                    case "System.Data.SqlTypes.SqlDouble"  : if (((SqlDouble)obj).IsNull)   return true; break;
                    case "System.Data.SqlTypes.SqlGuid"    : if (((SqlGuid)obj).IsNull)     return true; break;
                    case "System.Data.SqlTypes.SqlInt16"   : if (((SqlInt16)obj).IsNull)    return true; break;
                    case "System.Data.SqlTypes.SqlInt32"   : if (((SqlInt32)obj).IsNull)    return true; break;
                    case "System.Data.SqlTypes.SqlInt64"   : if (((SqlInt64)obj).IsNull)    return true; break;
                    case "System.Data.SqlTypes.SqlMoney"   : if (((SqlMoney)obj).IsNull)    return true; break;
                    case "System.Data.SqlTypes.SqlSingle"  : if (((SqlSingle)obj).IsNull)   return true; break;
                    case "System.Data.SqlTypes.SqlString"  : if (((SqlString)obj).IsNull)   return true; break;
                    case "System.Data.SqlTypes.SqlXml"     : if (((SqlXml)obj).IsNull)      return true; break;
                }


            return false;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Largest -->
        /// <summary>
        ///      Returns the largest item in the list
        /// </summary>
        /// <param name="list"></param>
        /// <returns></returns>
        public static int Largest(int[] list)
        {
            int max;
            max = Int32.MinValue;
            for (int i = 0; i < list.Length - 1; i++)
            {
                if (list[i] > max) max = list[i];
            }
            return max;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- SqlBool -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static SqlBoolean SqlBool(object obj, SqlBoolean defaultValue)
        {
            if (IsNull(obj))
            {
                return defaultValue;
            }
            else
            {
                Type type = obj.GetType();


                // ----------------------------------------------------------------------
                //  Common situations
                // ----------------------------------------------------------------------
                if (type == typeof(SqlBoolean) || type == typeof(Boolean) || type == typeof(bool))
                    return (SqlBoolean)obj;
                string str = obj.ToString();
                if (str == type.ToString())
                    throw new DataException("_.SqlBool - ToString() not defined in any useful way for " + type.ToString());


                // ----------------------------------------------------------------------
                //  Convert string
                // ----------------------------------------------------------------------
                return SqlBool(str, defaultValue);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- SqlBool -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="str"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static SqlBoolean SqlBool(string str, SqlBoolean defaultValue)
        {
            // --------------------------------------------------------------------------
            //  Common values
            // --------------------------------------------------------------------------
            if (string.IsNullOrEmpty(str))
                return defaultValue;
            try { SqlBoolean value = SqlBoolean.Parse(str); return value; }
            catch { }


            // --------------------------------------------------------------------------
            //  Additional approaches - 
            //  1/0, acceptable/unacceptable, complete/incomplete, correct/incorrect
            //  , engaged/disengaged, equal/unequal, good/bad, high/low, in/out
            //  , non-null/null, ok/not ok, open/closed, positive/negative
            //  , present/absent, right/wrong, true/false, yes/no
            //  etc.
            // --------------------------------------------------------------------------
            if (Regex.IsMatch(str, "^(accept|complet|engage|non-nul|nonnul|posit)", RegexOptions.IgnoreCase)) return true;
            if (Regex.IsMatch(str, "^(diseng|incompl|incorr|negat|unaccept)", RegexOptions.IgnoreCase)) return false;
            if (Regex.IsMatch(str, "^(correct|equal|good|hi|high|in|ok|open|present|right|true|yes)$", RegexOptions.IgnoreCase)) return true;
            if (Regex.IsMatch(str, "^(absent|bad|closed|false|lo|low|no|nil|not ok|null?|out|unequal|wrong)$", RegexOptions.IgnoreCase)) return false;


            // --------------------------------------------------------------------------
            //  letters numbers and symbols:
            //  Y/N, T/F, R/W, A/U, E/D, G/B, H/L, C/I, JMOPV/KQSXZ, +/-, &/|, 1-9/0, >/<
            // --------------------------------------------------------------------------
            if (Regex.IsMatch(str, "^[YTRAEGHCJMOPV]$", RegexOptions.IgnoreCase)) return true;  // AbCd EfGH iJklMn OPqRsT uVwxYz
            if (Regex.IsMatch(str, "^[NFWUDBLIKQSXZ]$", RegexOptions.IgnoreCase)) return false;
            if (Regex.IsMatch(str, "^[+1-9&>]$"     , RegexOptions.IgnoreCase)) return true;
            if (Regex.IsMatch(str, "^[1-9][0-9]+$", RegexOptions.IgnoreCase)) return true;
            if (Regex.IsMatch(str, "^[-0|<]$", RegexOptions.IgnoreCase)) return false;


            // --------------------------------------------------------------------------
            //  Give up
            // --------------------------------------------------------------------------
            return defaultValue;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- SqlXml -->
        /// <summary>
        ///      Example Usage:  SqlXml sx = SqlXml(tiny);
        /// </summary>
        /// <param name="str"></param>
        /// <returns></returns>
        public static SqlXml SqlXml(string str)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(str);
            XmlNodeReader rdr = new XmlNodeReader(doc);
            SqlXml xml = new SqlXml(rdr);
            return xml;
        }


        public static SqlXml ToSqlXml(string str)
        {
            Byte[] bytes = System.Text.Encoding.ASCII.GetBytes(str);
            MemoryStream stream = new MemoryStream(bytes);
            SqlXml xml = new SqlXml(stream);
            return xml;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StrValue -->
        /// <summary>
        ///      Converts an object to a string defaulting as specified if it is a null
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="defaultString"></param>
        /// <returns></returns>
        public static string StrValue(object obj, string defaultValue)
        {
            if (IsNull(obj)) return defaultValue;
            else
            {
                //if (obj.GetType() == typeof(Control))
                //    return ((Control)obj).Text;
                return obj.ToString();
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Time -->
        /// <summary>
        ///
        /// </summary>
        /// <param name="time"></param>
        /// <param name="defaultTime"></param>
        /// <returns></returns>
        public static TimeSpan Time(string str, TimeSpan defaultValue)
        {
            if (string.IsNullOrEmpty(str))
                return defaultValue;
            else
            {
                TimeSpan time;
                if (TimeSpan.TryParse(str, out time)) return time;
                else return defaultValue;
            }
        }

    }
}
