﻿using CHNw.CommonLib.Core;                        // for _
using CHNw.CommonLib.Diff;
using NUnit.Framework;                            // for testing
using NUnit.Framework.SyntaxHelpers;              // for testing
using System;
using System.Collections.Generic;                 // for List<>
using System.Text.RegularExpressions;                       /// for Regex

namespace DataWarehouse.LibTests
{
    // ------------------------------------------------------------------------------------------100
    /// <!-- DiffTests -->
    /// <summary>
    ///      The DiffTests class tests the classes in the Diff Library
    /// </summary>
    [TestFixture]
    public class DiffTests
    {
        private string _result;
        private static string _test1 = "\r\n((((((((((   ";
        private static string _test2 = " test   ))))))))))\r\n";


        // --------------------------------------------------------------------------------------100
        /// <!-- AllTests -->
        /// <summary>
        ///      Tests focused on the core classes: _, Date, Path, RandomSource
        /// </summary>
        public string AllTests()
        {
            return (Lcs_tests());
        }
        

        // --------------------------------------------------------------------------------------100
        /// <!-- _Method_tests -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        private string Lcs_tests()
        {
            _result = "";
            string result = "";


            // -------------------------------------------------------------------------90
            //  Method tests
            // -------------------------------------------------------------------------90
            LcsList_test            ();   result += _.Affix(_test1 + "LcsList"         + _test2, _result);
            LcsString_test          ();   result += _.Affix(_test1 + "LcsString"       + _test2, _result);


            return result;
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- LcsList_test -->
        /// <summary>
        ///      Tests least common string (for lists)
        /// </summary>
        [Test]
        private void LcsList_test()
        {
            List<string> list1 = new List<string>();
            List<string> list2;
            list1.Add("abc");
            list1.Add("def");
            //LcsList_testcase(list1, null        , 0); /// this is not supposed to work
            LcsList_testcase(list1, list1, 2, 0);

            list2 = new List<string>();
                              LcsList_testcase(list1, list2, 0, 2);
            list2.Add("ghi"); LcsList_testcase(list1, list2, 0, 3);
            list2.Add("abc"); LcsList_testcase(list1, list2, 1, 2);
            list2.Add("ghq"); LcsList_testcase(list1, list2, 1, 3);
            list2.Add("def"); LcsList_testcase(list1, list2, 2, 2);

            list2 = new List<string>();
            list2.Add(""   ); LcsList_testcase(list1, list2, 0, 3);
            list2.Add("def"); LcsList_testcase(list1, list2, 1, 2);
            list2.Add("abc"); LcsList_testcase(list1, list2, 1, 3); /// reverse order
            list2.Add("def"); LcsList_testcase(list1, list2, 2, 2);
            list2.Add("abc"); LcsList_testcase(list1, list2, 2, 3);

            list2 = new List<string>();
            list2.Add(null    ); LcsList_testcase(list1, list2, 0, 3);
            list2.Add("abcdef"); LcsList_testcase(list1, list2, 0, 4);

            _result = "ok";
        }
        private static void LcsList_testcase(List<string> list1, List<string> list2, int tgtSame, int tgtDiff)
        {
            DiffListLcs lcs = new DiffListLcs(list1, list2);
            List<string> list3 = lcs.CommonSequence();
            Assert.That(list3.Count == tgtSame);
            DiffList list4 = lcs.Diff();
            Assert.That(list4.ChangeCount == tgtDiff);
        }


        // --------------------------------------------------------------------------------------100
        /// <!-- LcsString_test -->
        /// <summary>
        ///      Tests least common string (for strings)
        /// </summary>
        [Test]
        public void LcsString_test()
        {
            ///                str1        str2        lcs      diff
            LcsString_testcase(null      , null      , ""     , ""                );
            LcsString_testcase(null      , ""        , ""     , ""                );
            LcsString_testcase(""        , null      , ""     , ""                );
            LcsString_testcase(""        , ""        , ""     , ""                );
            LcsString_testcase("a"       , ""        , ""     , "-a"              );
            LcsString_testcase("abc"     , ""        , ""     , "-a-b-c"          );
            LcsString_testcase("abc"     , "abc"     , "abc"  , ""                );
            LcsString_testcase(""        , "b"       , ""     , "+b"              );
            LcsString_testcase(""        , "abc"     , ""     , "+a+b+c"          );
            LcsString_testcase("a"       , "b"       , ""     , "-a+b"            );
            LcsString_testcase("abc"     , "def"     , ""     , "-a+d-b+e-c+f"    );
            LcsString_testcase("abc"     , "cba"     , "a"    , "+c+b-b-c"        );
            LcsString_testcase("abcs"    , "alphabet", "ab"   , "+l+p+h+a-c+e-s+t");
            LcsString_testcase("elephant", "alphabet", "lphat", "-e+a-e-n+b+e"    );

            _result = "ok";
        }
        private static void LcsString_testcase(string str1, string str2, string tgt, string diffTgt)
        {
            DiffString lcs = DiffString.GetLCS(str1, str2);

            string   str      = lcs.ToString();
            int[,]   matrix   = lcs.Table;
            int      xAxis    = matrix.GetLength(0);
            int      yAxis    = matrix.GetLength(1);
            int      length   = lcs.Length;
            string   sequence = lcs.Sequence();

            int len1   = _.Max("", str1).Length;
            int len2   = _.Max("", str2).Length;
            int expect = 3*(len1*len2 + len1 + len2) + 1;

            Assert.That(str.Length == expect);
            Assert.That(xAxis      == len1 + 1);
            Assert.That(yAxis      == len2 + 1);
            Assert.That(length     == tgt.Length);
            Assert.That(sequence   == tgt);

            string halfTgt = Regex.Replace(diffTgt, @"\-.", "");
            halfTgt = Regex.Replace(halfTgt, @"\+", "");
            string half = lcs.HalfDiff(1);
            Assert.That (half == halfTgt);
            string diff = lcs.Diff1to2();
            Assert.That(diff == diffTgt);
        }


    }
}
