// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;                                     // for Math

namespace InformationLib.Strings // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- StringPair -->
    /// <summary>
    ///      Contains a pair of strings usually for diff type comparisons
    /// </summary>
    public class StringPair
    {
        // ----------------------------------------------------------------------------------------
        //  Members and Properties
        // ----------------------------------------------------------------------------------------
        public string A          { get { return _strA;   } } private string _strA;
        public string B          { get { return _strB;   } } private string _strB;
        public char[] A_char     { get { return _charA;  } } private char[] _charA;
        public char[] B_char     { get { return _charB;  } } private char[] _charB;
        public int    A_count    { get { return _countA; } } private int    _countA;
        public int    B_count    { get { return _countB; } } private int    _countB;


        // ----------------------------------------------------------------------------------------
        //  Calculations
        // ----------------------------------------------------------------------------------------
        public int    LengthMin  { get { return Math.Min(_strA.Length, _strB.Length); } }
        public int    LengthMax  { get { return Math.Max(_strA.Length, _strB.Length); } }
        public int    LengthDiff { get { return Math.Abs(_strA.Length - _strB.Length); } }
        public bool   Same       { get { return (_strA == _strB); } }


        // ----------------------------------------------------------------------------------------
        //  Comparison results
        // ----------------------------------------------------------------------------------------
        public int    A_position  { get { return _positionA;   } } private int _positionA;
        public int    B_position  { get { return _positionB;   } } private int _positionB;
        public int    MatchLength { get { return _matchLength; } } private int _matchLength;
        public StringPair Before  { get { return new StringPair(A.Substring(0, A_position), B.Substring(0, B_position)); } }
        public StringPair Match   { get { return new StringPair(A.Substring(A_position, MatchLength), B.Substring(B_position, MatchLength)); } }
        public StringPair After   { get { return new StringPair(A.Substring(A_position + MatchLength), B.Substring(B_position + MatchLength)); } }


        // ----------------------------------------------------------------------------------------
        //  Constructor
        // ----------------------------------------------------------------------------------------
        public StringPair(string strA, string strB)
        {
            _strA   = strA;
            _strB   = strB;
            _charA  = strA.ToCharArray();
            _charB  = strB.ToCharArray();
            _countA = _charA.Length;
            _countB = _charB.Length;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- FindLongestMatch -->
        /// <summary>
        ///      Inefficiently finds the longest common string between two strings
        /// </summary>
        public void FindLongestMatch()
        {
            _matchLength = 0;
            for (int a = 0; a < _countA; ++a)
                for (int b = 0; b < _countB; ++b)
                {
                    int newLen = SamenessLength(a, b);
                    if (newLen > _matchLength)
                    {
                        _matchLength = newLen;
                        _positionA = a;
                        _positionB = b;
                    }
                }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- SamenessLength -->
        /// <summary>
        ///      Starting from a point in each string how long are they the same
        /// </summary>
        /// <param name="startA"></param>
        /// <param name="startB"></param>
        /// <returns></returns>
        private int SamenessLength(int startA, int startB)
        {
            int len;
            for (len = 0; startA+len < A_count && startB+len < B_count
                && A_char[startA+len] == B_char[startB+len]; ++len) { }
            return len;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            if (string.IsNullOrEmpty(_strA) && string.IsNullOrEmpty(_strB))
                return "";
            else return _strA + "\r\n" + _strB;
        }


    }
}
