// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;                                     // for DateTime, Guid
using System.Collections.Generic;                 // for Dictionary<,>
using System.Data;                                // for NoNullAllowedException
using System.IO;                                  // for File, StreamWriter, TextReader
using System.Text;                                // for UTF8Encoding
using System.Text.RegularExpressions;             // for Regex

namespace InformationLib.Strings // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- _ -->
    /// <summary>
    ///      The _ class is a place for static methods and properties oriented on:
    ///      1. string manipulations that I do not want to take up much code real estate
    ///      2. methods to be refactored into more reasonable places as they collect (see remarks)
    /// </summary>
    /// <remarks>
    ///      Graduates: (classes generally refactored from groups of methods that collected here)
    ///        for complex base conversions     see BigNumber
    ///        for complex date handling        see Date
    ///        for complex path manipulation    see PathSlicer
    ///        for complex form writing         see Plaster
    ///        for complex form reading         see Scrape
    ///        for complex exception handling   see Throw
    ///        for simple string parsing        see the Strings namespace
    ///        for simple type conversions      see TreatAs
    ///        for testing private methods      see ReflectorTest
    /// 
    ///      Patterns:
    ///        Singleton? - no - it's just a static class, there is no resource to manage
    /// 
    ///      This namespace is a 'primitive' so it should use nothing but System references
    /// </remarks>
    public static class _
    {

        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Concatenates two strings if neither is blank
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static string Affix(string str1, string str2)
        {
            if (!string.IsNullOrEmpty(str1)
                && !string.IsNullOrEmpty(str2))
                return str1 + str2;
            else
                return "";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Concatenates two strings if neither is blank
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="num"></param>
        /// <returns></returns>
        public static string Affix(string str, int num)
        {
            if (num >= 0)
                return str + num.ToString();
            else
                return "";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="delim"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static StringBuilder Affix(StringBuilder str1, string delim, string str2)
        {
            if (str1.Length > 0 && !string.IsNullOrEmpty(str2))
                str1.Append(delim);
            if (!string.IsNullOrEmpty(str2))
                str1.Append(str2);
            return str1;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Concatenates two strings if neither is blank
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="delim"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static string Affix(string str1, string delim, string str2)
        {
            if (!string.IsNullOrEmpty(str1) && !string.IsNullOrEmpty(str2))
                return str1 + delim + str2;
            else
                return str1 + str2;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Returns what strings contain content with delimiters as needed
        /// </summary>
        /// <param name="str1">string 1</param>
        /// <param name="d12">delimiter between string 1 and 2</param>
        /// <param name="str2">string 2</param>
        /// <param name="d23">delimiter between string 2 and 3</param>
        /// <param name="str3">string 3</param>
        /// <param name="d13">delimiter between string 1 and 3</param>
        /// <returns></returns>
        public static string Affix(string str1, string d12, string str2, string d23, string str3, string d13)
        {
            if (NO(str1) && NO(str2) && NO(str3)) return "";
            if (NO(str1) && NO(str2) && !NO(str3)) return str3;
            if (NO(str1) && !NO(str2) && NO(str3)) return str2;
            if (NO(str1) && !NO(str2) && !NO(str3)) return str2 + d23 + str3;
            if (!NO(str1) && NO(str2) && NO(str3)) return str1;
            if (!NO(str1) && NO(str2) && !NO(str3)) return str1 + d13 + str3;
            if (!NO(str1) && !NO(str2) && NO(str3)) return str1 + d12 + str2;
            return str1 + d12 + str2 + d23 + str3;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Affix -->
        /// <summary>
        ///      Returns what strings contain content with delimiters as needed
        /// </summary>
        /// <param name="str1">string 1</param>
        /// <param name="d12">delimiter between string 1 and 2</param>
        /// <param name="str2">string 2</param>
        /// <param name="d23">delimiter between string 2 and 3</param>
        /// <param name="str3">string 3</param>
        /// <param name="d34">delimiter between string 3 and 4</param>
        /// <param name="str4">string 4</param>
        /// <param name="d14">default delimiter</param>
        /// <returns></returns>
        public static string Affix(string str1, string d12, string str2, string d23, string str3
            , string d34, string str4, string d14)
        {
            if (NO(str1) && NO(str2) && NO(str3) && NO(str4)) return "";
            if (NO(str1) && NO(str2) && NO(str3) && !NO(str4)) return str4;
            if (NO(str1) && NO(str2) && !NO(str3) && NO(str4)) return str3;
            if (NO(str1) && NO(str2) && !NO(str3) && !NO(str4)) return str3 + d34 + str4;
            if (NO(str1) && !NO(str2) && NO(str3) && NO(str4)) return str2;
            if (NO(str1) && !NO(str2) && NO(str3) && !NO(str4)) return str2 + d14 + str4;
            if (NO(str1) && !NO(str2) && !NO(str3) && NO(str4)) return str2 + d23 + str3;
            if (NO(str1) && !NO(str2) && !NO(str3) && !NO(str4)) return str2 + d23 + str3 + d34 + str4;
            if (!NO(str1) && NO(str2) && NO(str3) && NO(str4)) return str1;
            if (!NO(str1) && NO(str2) && NO(str3) && !NO(str4)) return str1 + d14 + str4;
            if (!NO(str1) && NO(str2) && !NO(str3) && NO(str4)) return str1 + d14 + str3;
            if (!NO(str1) && NO(str2) && !NO(str3) && !NO(str4)) return str1 + d14 + str3 + d34 + str4;
            if (!NO(str1) && !NO(str2) && NO(str3) && NO(str4)) return str1 + d12 + str2;
            if (!NO(str1) && !NO(str2) && NO(str3) && !NO(str4)) return str1 + d12 + str2 + d14 + str4;
            if (!NO(str1) && !NO(str2) && !NO(str3) && NO(str4)) return str1 + d12 + str2 + d23 + str3;
            if (!NO(str1) && !NO(str2) && !NO(str3) && !NO(str4)) return str1 + d12 + str2 + d23 + str3 + d34 + str4;
            return "";
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Best -->
        /// <summary>
        ///      Returns the string that is not blank, preferring the first string
        /// </summary>
        /// <param name="primary"></param>
        /// <param name="secondary"></param>
        /// <returns></returns>
        public static string Best(string primary, string secondary)
        {
            if (string.IsNullOrEmpty(primary))
                return secondary;
            else
                return primary;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Collate -->
        /// <summary>
        ///      Collates a list of strings into a string given a joining string
        /// </summary>
        /// <param name="list">the list of strings to join</param>
        /// <param name="delim">the delimiter to join them with</param>
        /// <returns>a string from the items joined by the delimiter</returns>
        public static string Collate(List<string> list, string delim)
        {
            StringBuilder str = new StringBuilder();
            string d = "";
            foreach (string s in list) { str.Append(d + s); d = delim; }
            return str.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Collate -->
        /// <summary>
        ///      Returns a string of the contents of a dictionary collated in key order
        /// </summary>
        /// <param name="pos"></param>
        /// <param name="join"></param>
        /// <returns></returns>
        public static string Collate(Dictionary<char, int> hash, string join)
        {
            StringBuilder str = new StringBuilder();
            List<char> keys = KeysSorted(hash);
            string delim = "";
            foreach (char key in keys)
            {
                str.Append(delim).Append(hash[key]);
                delim = join;
            }
            return str.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- FileName -->
        /// <summary>
        ///      Gets a file name from a file path
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        public static string FileName(string filePath)
        {
            string fileName = Regex.Replace(filePath, "^.*[/\\\\]([^/\\\\]+)$", "$1");
            return fileName;
        }


        public static string Gap { get { return "\r\n\r\n"; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- GetListFromFile -->
        /// <summary>
        ///      Gets a list of the lines in a file
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        public static List<string> GetListFromFile(string path)
        {
            string str = GetStringFromFile(path, "");
            str = str.ToLower();
            char[] delimiter = { '\n', '\r' };
            List<string> name = new List<string>(str.Split(delimiter, StringSplitOptions.RemoveEmptyEntries));
            return name;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- GetStringFromFile -->
        /// <summary>
        ///      Gets a non-humongous string from a file
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="content"></param>
        public static string GetStringFromFile(string filePath, string defaultContent)
        {
            string content;
            TextReader tr;


            try
            {
                tr = new StreamReader(filePath);
                content = tr.ReadToEnd(); tr.Close();
            }
            catch { content = defaultContent; }


            return content;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- InstancesOf -->
        /// <summary>
        ///      Returns the nubmer of instances of a character in a string
        /// </summary>
        /// <param name="cha"></param>
        /// <param name="str"></param>
        /// <returns></returns>
        public static int InstancesOf(char cha, string str)
        {
            char[] array = str.ToCharArray();
            int count = 0;
            for (int i = 0; i < array.Length; ++i)
                if (array[i] == cha) ++count;
            return count;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- InTab -->
        /// <summary>
        ///      Converts an object to a seven character string for use in a tabbed list or table
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public static string InTab(object obj)
        { if (obj != null) return obj.ToString().PadLeft(6, ' ').Substring(0, 6); else return "      "; }


        // ----------------------------------------------------------------------------------------
        /// <!-- KeysSorted -->
        /// <summary>
        ///      Returns a sorted list of the keys of a dictionary
        /// </summary>
        /// <param name="hash"></param>
        /// <returns></returns>
        public static List<char> KeysSorted(Dictionary<char, int> hash)
        {
            List<char> keys = new List<char>(hash.Keys.Count);
            foreach (char key in hash.Keys)
                keys.Add(key);
            keys.Sort();
            return keys;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Max -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static string Max(string str1, string str2)
        {
            if (str1 == null) return str2;
            if (str2 == null) return str1;
            if (str1.Length > str2.Length) return str1;
            if (str2.Length > str1.Length) return str2;
            StringComparer sc = StringComparer.Create(System.Globalization.CultureInfo.InvariantCulture, false);
            int  c = sc.Compare(str1, str2);
            if (c == 0) return str1;
            return str2;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- MaxSeparation -->
        /// <summary>
        ///      Returns the longest distance between two instances of the character in a string
        /// </summary>
        /// <param name="cha"></param>
        /// <param name="str"></param>
        /// <returns></returns>
        public static int MaxSeparation(char cha, string str)
        {
            char[] array = str.ToCharArray();
            int count = 0;
            int maxCount = 0;
            for (int i = 0; i < array.Length; ++i)
                if (array[i] == cha) { maxCount = count; count = 0; }
                else count++;
            return maxCount;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- NearPath -->
        /// <summary>
        ///      Returns a path close to the path input
        /// </summary>
        /// <param name="whereFoundPath"></param>
        /// <returns></returns>
        public static string NearPath(string whereFoundPath, string suffixFrom, string suffixTo)
        {
            string filePath = Regex.Replace(whereFoundPath, suffixFrom + "$", suffixTo);
            if (filePath == whereFoundPath)
                filePath = "..\\" + filePath;
            return filePath;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- NO -->
        /// <summary>
        ///      String is null or empty (syntactic candy)
        /// </summary>
        /// <param name="str1"></param>
        /// <returns></returns>
        private static bool NO(string str)
        {
            return string.IsNullOrEmpty(str);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Openface -->
        /// <summary>
        ///      Concatenates a delimiter between two strings if either is not ""
        /// </summary>
        /// <param name="str1"></param>
        /// <param name="delim"></param>
        /// <param name="str2"></param>
        /// <returns></returns>
        public static string Openface(string str1, string delim, string str2)
        {
            if (string.IsNullOrEmpty(str1)
                && string.IsNullOrEmpty(str2))
                return "";
            else
                return str1 + delim + str2;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- PrettyXml -->
        /// <summary>
        ///      Returns a pretty version of an xml string without changing the content,
        ///      adds and removes white space between the tags
        /// </summary>
        /// <param name="xml"></param>
        /// <param name="indent">number of spaces per indentation</param>
        /// <param name="cr">preferred carriage return, defaults to "\r\n"</param>
        /// <returns></returns>
        public static string PrettyXml(string xml, int indent, string cr)
        {
            // -------------------------------------------------------------------------90
            //  Split xml into tag lines
            // -------------------------------------------------------------------------90
            xml = Regex.Replace(xml, ">[\r\n\t ]*<", ">\r<");
            char[] separators = { '\r' };
            string[] line = xml.Split(separators, StringSplitOptions.RemoveEmptyEntries);



            StringBuilder xml2 = new StringBuilder();
            if (string.IsNullOrEmpty(cr)) cr = "\r\n";
            int depth = 0;
            int curr = 0;
            List<string> stack = new List<string>();
            for (int i = 0; i < line.Length; ++i)
            {
                // ---------------------------------------------------------------------90
                //  Calculate line
                // ---------------------------------------------------------------------90
                if      (Regex.IsMatch(line[i], "^</[^<>]+>$")) { depth--; curr = depth; } // depth decreases are immediate
                else if (Regex.IsMatch(line[i], "^<[^</>][^<>]*[^</>]>$")) depth++;  // handle 2+ character tags
                else if (Regex.IsMatch(line[i], "^<[^</>]>$")) depth++;  // handle single character tags
                xml2.Append(cr.PadRight(cr.Length + indent * curr) + line[i]);
                curr = depth;  // depth increases are delayed
            }
            string xml3 = Regex.Replace(xml2.ToString(), "^[\r\n]*", "");
            return xml3;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomId -->
        /// <summary>
        ///      Builds an ID where the year, month, and day each consist of one letter
        /// </summary>
        /// <remarks>
        ///      First Letter:
        ///        A 2000,2025,2050,2075      B 2001,2026,2051,2076     C 2002, 2027, 2052, 2077
        ///        ...
        ///        Y 2024,2049,2074,2099
        /// 
        ///      Second Letter: (the 2nd letter of the pair is for days after the 26th)
        ///        AB January    CD February   EF March      GH April      IJ May        KL June
        ///        MN July       OP August     QR Sept       ST Oct        UV Nov        WX Dec
        /// 
        ///      Third Letter (16):
        ///        A 1st,17th    B 2nd,18th    C 3rd,19th    D 4th,20th    E 5th, 21sh
        ///        F 6th,22nd    G 7th,23rd    H 8th,24th    I 9th,25th    J 10th,26th
        ///        K 11th,27th   L 12th,28th   M 13th,29th   N 14th,30th   O 15th,31st   P 16th   
        /// 
        ///      Third Letter (27):
        ///        A 1st,27th    B 2nd,28th    C 3rd,29th    D 4th,30th    E 5th, 31st
        ///        F 6th     G 7th     H 8th     I 9th     J 10th    K 11th    L 12th    M 13th
        ///        N 14th    O 15th    P 16th    Q 17th    R 18th    S 19th    T 20th    U 21st
        ///        V 22nd    W 23rd    X 24th    Y 25th    Z 26th
        /// 
        ///      Fourth Letter:
        ///        A 0:00    B 1:00    C 2:00    D 3:00    E 4:00
        ///        F 5:00    G 6:00    H 7:00    I 8:00    J 9:00    K 10:00   L 11:00   M 12:00
        ///        N 13:00   O 14:00   P 15:00   Q 16:00   R 17:00   S 18:00   T 19:00   U 20:00
        ///        V 21:00   W 22:00   X 23:00
        /// 
        ///      So 3/17/2011 3:30 PM would be  LEQP , so would 3/17/2036 3:10 PM
        /// </remarks>
        /// <param name="prefix"></param>
        /// <param name="date"></param>
        /// <param name="numNumerals"></param>
        /// <param name="join">string to join various the sub parts</param>
        /// <returns></returns>
        public static string RandomId_EmbedDate(string prefix, DateTime date, string join, int numDigits)
        {
            int wrapday = 16;  /// must be between 16 and 26
            if (string.IsNullOrEmpty(join)) join = "";
            string year  = "" + (char)(65 + date.Year % 25);
            string month = "" + (char)(63 + date.Month * 2 + ((date.Day - 1) / wrapday)); // two letters per month
            string day   = "" + (char)(65 + (date.Day-1) % wrapday);                // the end of the month wraps back to A,B,C...
            string hour  = "" + (char)(65 + date.Hour);
            string id = year + month + day + hour;
            if (!string.IsNullOrEmpty(prefix)) id = prefix + id;
            if (numDigits > 0) id = id + join + _.RandomNumericId(numDigits);
            return id;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomNumericId -->
        /// <summary>
        ///      Uses Guids to create random numeric id's
        /// </summary>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string RandomNumericId(int length)
        {
            string id = "";
            int subLen = 17;
            while (length > 0)
            {
                if (length > subLen) id += RandomNumericId(subLen, subLen);
                else id += RandomNumericId(length, subLen);
                length -= subLen;
            }

            return id;
        }
        /// <summary>creates a 'random' numeric ID up to maxLen digits long using a Guid</summary>
        private static string RandomNumericId(int length, int maxLen)
        {
            if (length > maxLen) length = maxLen;
            long num = BitConverter.ToInt64(Guid.NewGuid().ToByteArray(), 0);
            string id = num.ToString().PadLeft(length + 10, '0');
            id = id.Substring(id.Length - length);
            return id;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RenameNamespace -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="xml"></param>
        /// <param name="ns1"></param>
        /// <param name="ns2"></param>
        /// <param name="ns3"></param>
        /// <returns></returns>
        public static string RenameNamespace(string xml, string renameFrom, string renameTo, string ns3)
        {
            xml = Regex.Replace(xml, renameTo, ns3);
            xml = Regex.Replace(xml, renameFrom, renameTo);
            return xml;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Reverse -->
        /// <summary>
        ///      Reverses a string
        /// </summary>
        /// <param name="str"></param>
        /// <returns></returns>
        public static string Reverse(string str)
        {
            if (str == null) return null;
            if (str.Length <= 1) return str;
            char[] input = str.ToCharArray();
            string output = "";
            for (int i = input.Length - 1; i >= 0; --i)
                output += input[i];
            return output;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- SetLength -->
        /// <summary>
        ///      Returns a string exactly as long as specified
        /// </summary>
        /// <param name="str"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static string SetLength(int width, string str)
        {
            if (str == null) str = "";
            str = str.PadRight(width+1, ' ').Substring(0,width);
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToggleCase -->
        /// <summary>
        ///      Toggles a character's case
        /// </summary>
        /// <param name="c"></param>
        /// <returns></returns>
        public static char ToggleCase(char c)
        {
            char c2 = c;
            if (Char.IsUpper(c)) c2 = Char.ToLower(c);
            if (Char.IsLower(c)) c2 = Char.ToUpper(c);
            return c2;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Underline -->
        /// <summary>
        ///      Underlines a string
        /// </summary>
        /// <param name="title"></param>
        /// <param name="ul"></param>
        /// <returns></returns>
        public static string Underline(string title, char ul)
        {
            string str = title + "\r\n" + ul.ToString().PadRight(title.Length, ul);
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- UTF8ByteArrayToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="characters"></param>
        /// <returns></returns>
        public static String UTF8ByteArrayToString(Byte[] characters)
        {

            UTF8Encoding encoding = new UTF8Encoding();
            String constructedString = encoding.GetString(characters);
            return (constructedString);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StringToUTF8ByteArray -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="pXmlString"></param>
        /// <returns></returns>
        public static Byte[] StringToUTF8ByteArray(String pXmlString)
        {
            UTF8Encoding encoding = new UTF8Encoding();
            Byte[] byteArray = encoding.GetBytes(pXmlString);
            return byteArray;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- WriteStringToFile -->
        /// <summary>
        ///      Writes a non-humongous string to a file
        /// </summary>
        /// <remarks>Shouldn't this be done in another class?</remarks>
        /// <param name="filePath"></param>
        /// <param name="content"></param>
        public static void WriteStringToFile(string filePath, string content, bool append)
        {
            if (content == null)
            {
                throw new NoNullAllowedException("WriteStringToFile error - content is null");
            }
            else
            {
                string[] lines = new string[1];
                StreamWriter fileWriter = null;
                lines[0] = content;


                if (filePath == null || filePath.Length == 0) return;
                if (lines == null || lines.Length == 0) return;


                try
                {
                    if (append)
                        fileWriter = File.AppendText(filePath);
                    else
                        fileWriter = File.CreateText(filePath);
                    foreach (string line in lines) fileWriter.Write(line);
                }
                catch { throw new IOException("Unable to write string to file " + filePath); }
                finally { if (fileWriter != null)  fileWriter.Close(); }
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- BreakPoint -->
        /// <summary>
        ///      This is a no-op method to be used to provide a place to set breaks to
        /// </summary>
        public static void BreakPoint()
        {
        }


        public static void Pause()
        {
        }

    }
}
